"""Text features transformers."""

import gc
import logging
import os
import pickle

from copy import copy
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union


try:
    import gensim
except:
    import warnings

    warnings.warn("'gensim' - package isn't installed")

import numpy as np
import pandas as pd
import torch

from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import SGDClassifier
from sklearn.linear_model import SGDRegressor

from ..dataset.base import LAMLDataset
from ..dataset.np_pd_dataset import CSRSparseDataset
from ..dataset.np_pd_dataset import NumpyDataset
from ..dataset.np_pd_dataset import PandasDataset
from ..dataset.roles import NumericRole
from ..dataset.roles import TextRole
from ..text.dl_transformers import BOREP
from ..text.dl_transformers import BertEmbedder
from ..text.dl_transformers import DLTransformer
from ..text.dl_transformers import RandomLSTM
from ..text.embed_dataset import BertDataset
from ..text.embed_dataset import EmbedDataset
from ..text.tokenizer import BaseTokenizer
from ..text.tokenizer import SimpleEnTokenizer
from ..text.utils import get_textarr_hash
from ..text.weighted_average_transformer import WeightedAverageTransformer
from .base import LAMLTransformer


logger = logging.getLogger(__name__)

NumpyOrPandas = Union[NumpyDataset, PandasDataset]
NumpyOrSparse = Union[NumpyDataset, CSRSparseDataset]

model_by_name = {
    "random_lstm": {
        "model": RandomLSTM,
        "model_params": {
            "embed_size": 300,
            "hidden_size": 256,
            "pooling": "mean",
            "num_layers": 1,
        },
        "dataset": EmbedDataset,
        "dataset_params": {
            "embedding_model": None,
            "max_length": 200,
            "embed_size": 300,
        },
        "loader_params": {"batch_size": 1024, "shuffle": False, "num_workers": 4},
        "embedding_model_params": {},
    },
    "random_lstm_bert": {
        "model": RandomLSTM,
        "model_params": {
            "embed_size": 768,
            "hidden_size": 256,
            "pooling": "mean",
            "num_layers": 1,
        },
        "dataset": BertDataset,
        "dataset_params": {"max_length": 256, "model_name": "bert-base-cased"},
        "loader_params": {"batch_size": 320, "shuffle": False, "num_workers": 4},
        "embedding_model": BertEmbedder,
        "embedding_model_params": {"model_name": "bert-base-cased", "pooling": "none"},
    },
    "borep": {
        "model": BOREP,
        "model_params": {
            "embed_size": 300,
            "proj_size": 300,
            "pooling": "mean",
            "max_length": 200,
            "init": "orthogonal",
            "pos_encoding": False,
        },
        "dataset": EmbedDataset,
        "dataset_params": {
            "embedding_model": None,
            "max_length": 200,
            "embed_size": 300,
        },
        "loader_params": {"batch_size": 1024, "shuffle": False, "num_workers": 4},
        "embedding_model_params": {},
    },
    "pooled_bert": {
        "model": BertEmbedder,
        "model_params": {"model_name": "bert-base-cased", "pooling": "mean"},
        "dataset": BertDataset,
        "dataset_params": {"max_length": 256, "model_name": "bert-base-cased"},
        "loader_params": {"batch_size": 320, "shuffle": False, "num_workers": 4},
        "embedding_model_params": {},
    },
    "wat": {
        "embedding_model": None,
        "embed_size": 300,
        "weight_type": "idf",
        "use_svd": True,
    },
}


def oof_task_check(dataset: LAMLDataset):
    """Check if task is binary or regression.

    Args:
        dataset: Dataset to check.

    """
    task = dataset.task
    assert task.name in [
        "binary",
        "reg",
    ], "Only binary and regression tasks supported in this transformer"


def text_check(dataset: LAMLDataset):
    """Check if all passed vars are text.

    Args:
        dataset: LAMLDataset to check.

    Raises:
         AssertionError: If non-text features are present.

    """
    roles = dataset.roles
    features = dataset.features
    for f in features:
        assert roles[f].name == "Text", "Only text accepted in this transformer"


# TODO: combine TunableTransformer with LAMLTransformer class?
class TunableTransformer(LAMLTransformer):
    """Base class for ML transformers.

    Assume that parameters my set before training.

    Args:
        default_params: algo hyperparams.
        freeze_defaults:
            - ``True`` :  params may be rewritten depending on dataset.
            - ``False``:  params may be changed only manually or with tuning.

    """

    _default_params: dict = {}
    _params: dict = None

    @property
    def params(self) -> dict:
        """Parameters.

        Returns:
            Dict.

        """
        if self._params is None:
            self._params = copy(self.default_params)
        return self._params

    @params.setter
    def params(self, new_params: dict):
        assert isinstance(new_params, dict)
        self._params = {**self.params, **new_params}

    def init_params_on_input(self, dataset: NumpyOrPandas) -> dict:
        """Init params depending on input data.

        Args:
            dataset: Dataset.

        Returns:
            Dict with model hyperparameters.

        """
        return self.params

    def __init__(self, default_params: Optional[dict] = None, freeze_defaults: bool = True):
        self.task = None

        self.freeze_defaults = freeze_defaults
        if default_params is None:
            default_params = {}

        self.default_params = {**self._default_params, **default_params}


class TfidfTextTransformer(TunableTransformer):
    """Simple Tfidf vectorizer.

    Args:
        default_params: algo hyperparams.
        freeze_defaults: Flag.
        subs: Subsample to calculate freqs. If ``None`` - full data.
        random_state: Random state to take subsample.

    Note:
        The behaviour of `freeze_defaults`:

        - ``True`` :  params may be rewritten depending on dataset.
        - ``False``:  params may be changed only
            manually or with tuning.

    """

    _fit_checks = (text_check,)
    _transform_checks = ()
    _fname_prefix = "tfidf"
    _default_params = {
        "min_df": 5,
        "max_df": 1.0,
        "max_features": 30_000,
        "ngram_range": (1, 1),
        "analyzer": "word",
        "dtype": np.float32,
    }

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def __init__(
        self,
        default_params: Optional[dict] = None,
        freeze_defaults: bool = True,
        subs: Optional[int] = None,
        random_state: int = 42,
    ):
        super().__init__(default_params, freeze_defaults)
        self.subs = subs
        self.random_state = random_state
        self.vect = TfidfVectorizer
        self.dicts = {}

    def init_params_on_input(self, dataset: NumpyOrPandas) -> dict:
        """Get transformer parameters depending on dataset parameters.

        Args:
            dataset: Dataset used for model parameters initialization.

        Returns:
            Parameters of model.

        """
        # TODO: use features_num
        suggested_params = copy(self.default_params)
        if self.freeze_defaults:
            # if user change defaults manually - keep it
            return suggested_params
        rows_num = len(dataset.data)
        if rows_num > 50_000:
            suggested_params["min_df"] = 25

        return suggested_params

    def fit(self, dataset: NumpyOrPandas):
        """Fit tfidf vectorizer.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            self.

        """
        # set transformer names and add checks
        for check_func in self._fit_checks:
            check_func(dataset)
        # set transformer features

        # convert to accepted dtype and get attributes
        if self._params is None:
            self.params = self.init_params_on_input(dataset)
        dataset = dataset.to_pandas()
        df = dataset.data

        # fit ...
        if self.subs is not None and df.shape[0] >= self.subs:
            subs = df.sample(n=self.subs, random_state=self.random_state)
        else:
            subs = df

        feats = []
        for n, i in enumerate(subs.columns):
            vect = self.vect(**self.params)
            vect.fit(subs[i].fillna("").astype(str))
            features = list(
                np.char.array([self._fname_prefix + "_"])
                + np.arange(len(vect.vocabulary_)).astype(str)
                + np.char.array(["__" + i])
            )
            self.dicts[i] = {"vect": vect, "feats": features}
            feats.extend(features)
        self._features = feats
        return self

    def transform(self, dataset: NumpyOrPandas) -> CSRSparseDataset:
        """Transform text dataset to sparse tfidf representation.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            Sparse dataset with encoded text.

        """
        # checks here
        super().transform(dataset)
        # convert to accepted dtype and get attributes
        dataset = dataset.to_pandas()
        df = dataset.data

        # transform
        roles = NumericRole()
        outputs = []
        for n, i in enumerate(df.columns):
            new_arr = self.dicts[i]["vect"].transform(df[i].fillna("").astype(str))
            output = dataset.empty().to_numpy().to_csr()
            output.set_data(new_arr, self.dicts[i]["feats"], roles)
            outputs.append(output)
        # create resulted
        return dataset.empty().to_numpy().to_csr().concat(outputs)


class TokenizerTransformer(LAMLTransformer):
    """Simple tokenizer transformer.

    Args:
        tokenizer: text tokenizer.

    """

    _fit_checks = (text_check,)
    _transform_checks = ()
    _fname_prefix = "tokenized"

    def __init__(self, tokenizer: BaseTokenizer = SimpleEnTokenizer()):
        self.tokenizer = tokenizer

    def transform(self, dataset: NumpyOrPandas) -> PandasDataset:
        """Transform text dataset to tokenized text dataset.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            Pandas dataset with tokenized text.

        """
        # checks here
        super().transform(dataset)
        # convert to accepted dtype and get attributes
        dataset = dataset.to_pandas()
        df = dataset.data

        # transform
        roles = TextRole()
        outputs = []
        for n, i in enumerate(df.columns):
            pred = np.array(self.tokenizer.tokenize(df[i].fillna("").astype(str).tolist()))
            new_df = pd.DataFrame(pred, columns=[self._fname_prefix + "__" + i])
            outputs.append(new_df)
        # create resulted
        output = dataset.empty().to_pandas()
        output.set_data(pd.concat(outputs, axis=1), None, {feat: roles for feat in self.features})
        return output


class OneToOneTransformer(TunableTransformer):
    """Out-of-fold sgd model prediction to reduce dimension of encoded text data."""

    _fit_checks = (oof_task_check,)
    _transform_checks = ()
    _fname_prefix = "sgd_oof"
    _default_params = {"alpha": 0.0001, "max_iter": 1, "loss": "log"}

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def init_params_on_input(self, dataset: NumpyOrPandas) -> dict:
        """Get model parameters depending on dataset parameters.

        Args:
            dataset: NumpyOrPandas.

        Returns:
            Parameters of model.

        """
        # TODO: use features_num
        suggested_params = copy(self.default_params)

        self.task = dataset.task.name
        if self.task != "binary":
            suggested_params["loss"] = "squared_loss"

            algo = SGDRegressor
        else:
            algo = SGDClassifier

        self.algo = algo

        return suggested_params

    def __init__(self, default_params: Optional[int] = None, freeze_defaults: bool = False):
        super().__init__(default_params, freeze_defaults)
        """

        Args:
            default_params: Algo hyperparams.
            freeze_defaults:
                - ``True`` :  params may be rewritten depending on dataset.
                - ``False``:  params may be changed only manually or with tuning.
            subs: Subsample to calculate freqs. If None - full data.

        """

    def fit(self, dataset: NumpyOrPandas):
        """Apply fit transform.

        Args:
            dataset: Pandas or Numpy dataset of encoded text features.

        Returns:
            self.

        """
        for check_func in self._fit_checks:
            check_func(dataset)
        self.fit(dataset)

        for check_func in self._transform_checks:
            check_func(dataset)

        return self.transform(dataset)

    def fit_transform(self, dataset: NumpyOrPandas) -> NumpyDataset:
        """Fit and predict out-of-fold sgd model.

        Args:
            dataset: Pandas or Numpy dataset of encoded text features.

        Returns:
            Numpy dataset with out-of-fold model prediction.

        """
        # set transformer names and add checks
        for check_func in self._fit_checks:
            check_func(dataset)

        if self._params is None:
            self.params = self.init_params_on_input(dataset)
        dataset = dataset.to_numpy()
        data = dataset.data

        target = dataset.target.astype(np.int32)

        folds = dataset.folds
        n_folds = folds.max() + 1

        self.models = []

        oof_feats = np.zeros(len(data), dtype=np.float32)

        for n in range(n_folds):
            algo = self.algo(**self.params)
            algo.fit(data[folds != n], target[folds != n])
            if self.task == "binary":
                pred = algo.predict_proba(data[folds == n])[:, 1]
            else:
                pred = algo.predict(data[folds == n])
            oof_feats[folds == n] = pred

            self.models.append(deepcopy(algo))

        orig_name = dataset.features[0].split("__")[-1]
        self._features = [self._fname_prefix + "__" + orig_name]
        output = dataset.empty()
        self.output_role = NumericRole(np.float32, prob=output.task.name == "binary")
        output.set_data(oof_feats[:, np.newaxis], self.features, self.output_role)

        return output

    def transform(self, dataset: NumpyOrPandas) -> NumpyDataset:
        """Transform dataset to out-of-fold model-based encoding.

        Args:
            dataset: Pandas or Numpy dataset of encoded text features.

        Returns:
            Numpy dataset with out-of-fold model prediction.

        """
        # checks here
        super().transform(dataset)
        # convert to accepted dtype and get attributes
        dataset = dataset.to_numpy()
        data = dataset.data

        # transform
        out = np.zeros(len(data), dtype=np.float32)
        for n, model in enumerate(self.models):
            if self.task == "binary":
                pred = model.predict_proba(data)[:, 1]
            else:
                pred = model.predict(data)
            out += pred

        out /= len(self.models)
        # create resulted
        output = dataset.empty()
        output.set_data(out[:, np.newaxis], self.features, self.output_role)

        return output


class ConcatTextTransformer(LAMLTransformer):
    """Concat text features transformer.

    Args:
        special_token: Add special token between columns.

    """

    _fit_checks = (text_check,)
    _transform_checks = ()
    _fname_prefix = "concated"

    def __init__(self, special_token: str = " [SEP] "):
        self.special_token = special_token

    def transform(self, dataset: NumpyOrPandas) -> PandasDataset:
        """Transform text dataset to one text column.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            Pandas dataset with one text column.

        """
        # checks here
        super().transform(dataset)
        # convert to accepted dtype and get attributes
        dataset = dataset.to_pandas()
        df = dataset.data

        # transform
        roles = TextRole()
        new_df = pd.DataFrame(
            df[df.columns].fillna("").astype(str).apply(f"{self.special_token}".join, axis=1),
            columns=[self._fname_prefix + "__" + "__".join(df.columns)],
        )
        # create resulted
        output = dataset.empty().to_pandas()
        output.set_data(new_df, None, {feat: roles for feat in new_df.columns})
        return output


class AutoNLPWrap(LAMLTransformer):
    """Calculate text embeddings.

    Args:
        model_name: Method for aggregating word embeddings
            into sentence embedding.
        transformer_params: Aggregating model parameters.
        embedding_model: Word level embedding model with dict
            interface or path to gensim fasttext model.
        cache_dir: If ``None`` - do not cache transformed datasets.
        bert_model: Name of HuggingFace transformer model.
        subs: Subsample to calculate freqs. If None - full data.
        multigpu: Use Data Parallel.
        random_state: Random state to take subsample.
        train_fasttext: Train fasttext.
        fasttext_params: Fasttext init params.
        fasttext_epochs: Number of epochs to train.
        verbose: Verbosity.
        device: Torch device or str.
        **kwargs: Unused params.

    """

    _fit_checks = (text_check,)
    _transform_checks = ()
    _fname_prefix = "emb"
    fasttext_params = {"vector_size": 64, "window": 3, "min_count": 1}
    _names = {"random_lstm", "random_lstm_bert", "pooled_bert", "wat", "borep"}
    _trainable = {"wat", "borep", "random_lstm"}

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def __init__(
        self,
        model_name: str,
        embedding_model: Optional[str] = None,
        cache_dir: str = "./cache_NLP",
        bert_model: Optional[str] = None,
        transformer_params: Optional[Dict] = None,
        subs: Optional[int] = None,
        multigpu: bool = False,
        random_state: int = 42,
        train_fasttext: bool = False,
        fasttext_params: Optional[Dict] = None,
        fasttext_epochs: int = 2,
        sent_scaler: Optional[str] = None,
        verbose: bool = False,
        device: Any = "0",
        **kwargs: Any,
    ):
        if train_fasttext:
            assert model_name in self._trainable, f"If train fasstext then model must be in {self._trainable}"

        assert model_name in self._names, f"Model name must be one of {self._names}"
        self.device = device
        self.multigpu = multigpu
        self.cache_dir = cache_dir
        self.random_state = random_state
        self.subs = subs
        self.train_fasttext = train_fasttext
        self.fasttext_epochs = fasttext_epochs
        if fasttext_params is not None:
            self.fasttext_params.update(fasttext_params)
        self.dicts = {}
        self.sent_scaler = sent_scaler
        self.verbose = verbose
        self.model_name = model_name
        self.transformer_params = model_by_name[self.model_name]
        if transformer_params is not None:
            self.transformer_params.update(transformer_params)

        self._update_bert_model(bert_model)
        if embedding_model is not None:
            if isinstance(embedding_model, str):
                try:
                    embedding_model = gensim.models.FastText.load(embedding_model)
                except:
                    try:
                        embedding_model = gensim.models.FastText.load_fasttext_format(embedding_model)
                    except:
                        embedding_model = gensim.models.KeyedVectors.load(embedding_model)

            self.transformer_params = self._update_transformers_emb_model(self.transformer_params, embedding_model)

        else:

            self.train_fasttext = self.model_name in self._trainable

        if self.model_name == "wat":
            self.transformer = WeightedAverageTransformer
        else:
            self.transformer = DLTransformer

    def _update_bert_model(self, bert_model: str):
        if bert_model is not None:
            if "dataset_params" in self.transformer_params:
                self.transformer_params["dataset_params"]["model_name"] = bert_model
            if "embedding_model_params" in self.transformer_params:
                self.transformer_params["embedding_model_params"]["model_name"] = bert_model
            if "model_params" in self.transformer_params:
                self.transformer_params["model_params"]["model_name"] = bert_model
        return self

    def _update_transformers_emb_model(
        self, params: Dict, model: Any, emb_size: Optional[int] = None
    ) -> Dict[str, Any]:
        if emb_size is None:
            try:
                # Gensim checker [1]
                emb_size = model.vector_size
            except:
                try:
                    # Gensim checker[2]
                    emb_size = model.vw.vector_size
                except:
                    try:
                        # Natasha checker
                        emb_size = model[model.vocab.words[0]].shape[0]
                    except:
                        try:
                            # Dict of embeddings checker
                            emb_size = next(iter(model.values())).shape[0]
                        except:
                            raise ValueError("Unrecognized embedding dimension, please specify it in model_params")
        try:
            model = model.wv
        except:
            pass

        if self.model_name == "wat":
            params["embed_size"] = emb_size
            params["embedding_model"] = model
        elif self.model_name in {"random_lstm", "borep"}:
            params["dataset_params"]["embedding_model"] = model
            params["dataset_params"]["embed_size"] = emb_size
            params["model_params"]["embed_size"] = emb_size

        return params

    def fit(self, dataset: NumpyOrPandas):
        """Fit chosen transformer and create feature names.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            self.

        """
        for check_func in self._fit_checks:
            check_func(dataset)

        if self.cache_dir is not None:
            if not os.path.exists(self.cache_dir):
                os.makedirs(self.cache_dir)
        # set transformer features

        # convert to accepted dtype and get attributes
        dataset = dataset.to_pandas()
        df = dataset.data

        # fit ...
        if self.subs is not None and df.shape[0] >= self.subs:
            subs = df.sample(n=self.subs, random_state=self.random_state)
        else:
            subs = df

        names = []
        for n, i in enumerate(subs.columns):
            transformer_params = deepcopy(self.transformer_params)
            if self.train_fasttext:
                embedding_model = gensim.models.FastText(**self.fasttext_params)
                common_texts = [i.split(" ") for i in subs[i].values]
                embedding_model.build_vocab(corpus_iterable=common_texts)
                embedding_model.train(
                    corpus_iterable=common_texts,
                    total_examples=len(common_texts),
                    epochs=self.fasttext_epochs,
                )
                transformer_params = self._update_transformers_emb_model(transformer_params, embedding_model)

            transformer = self.transformer(
                verbose=self.verbose,
                device=self.device,
                multigpu=self.multigpu,
                **transformer_params,
            )
            emb_name = transformer.get_name()
            emb_size = transformer.get_out_shape()

            feats = [self._fname_prefix + "_" + emb_name + "_" + str(x) + "__" + i for x in range(emb_size)]

            self.dicts[i] = {
                "transformer": deepcopy(transformer.fit(subs[i])),
                "feats": feats,
            }
            names.extend(feats)
            logger.info3(f"Feature {i} fitted")

            del transformer
            gc.collect()
            torch.cuda.empty_cache()

        self._features = names
        return self

    def transform(self, dataset: NumpyOrPandas) -> NumpyOrPandas:
        """Transform tokenized dataset to text embeddings.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            Numpy dataset with text embeddings.

        """
        # checks here
        super().transform(dataset)
        # convert to accepted dtype and get attributes
        dataset = dataset.to_pandas()
        df = dataset.data

        # transform
        roles = NumericRole()
        outputs = []
        for n, i in enumerate(df.columns):
            if self.cache_dir is not None:
                full_hash = get_textarr_hash(df[i]) + get_textarr_hash(self.dicts[i]["feats"])
                fname = os.path.join(self.cache_dir, full_hash + ".pkl")
                if os.path.exists(fname):
                    logger.info3(f"Load saved dataset for {i}")
                    with open(fname, "rb") as f:
                        new_arr = pickle.load(f)
                else:
                    new_arr = self.dicts[i]["transformer"].transform(df[i])
                    with open(fname, "wb") as f:
                        pickle.dump(new_arr, f)
            else:
                new_arr = self.dicts[i]["transformer"].transform(df[i])

            output = dataset.empty().to_numpy()
            output.set_data(new_arr, self.dicts[i]["feats"], roles)
            outputs.append(output)
            logger.info3(f"Feature {i} transformed")
        # create resulted
        dataset = dataset.empty().to_numpy().concat(outputs)
        # instance-wise sentence embedding normalization
        dataset.data = dataset.data / self._sentence_norm(dataset.data, self.sent_scaler)

        return dataset

    @staticmethod
    def _sentence_norm(x: np.ndarray, mode: Optional[str] = None) -> Union[np.ndarray, float]:
        """Get sentence embedding norm."""
        if mode == "l2":
            return ((x ** 2).sum(axis=1, keepdims=True)) ** 0.5
        elif mode == "l1":
            return np.abs(x).sum(axis=1, keepdims=True)
        if mode is not None:
            logger.info2("Unknown sentence scaler mode: sent_scaler={}, " "no normalization will be used".format(mode))
        return 1



"""Basic classes for transformers."""

from copy import deepcopy
from typing import Callable
from typing import List
from typing import Sequence
from typing import Type
from typing import Union

import numpy as np
import pandas as pd

from ..dataset.base import LAMLDataset
from ..dataset.base import RolesDict
from ..dataset.roles import ColumnRole
from ..dataset.utils import concatenate


# TODO: From func transformer

Roles = Union[Sequence[ColumnRole], ColumnRole, RolesDict, None]


class LAMLTransformer:
    """Base class for transformer method (like sklearn, but works with datasets)."""

    _fname_prefix = None
    _fit_checks = ()
    _transform_checks = ()

    @property
    def features(self) -> List[str]:
        """Get name of the features, that will be generated after transform.

        Returns:
            List of new names.

        """
        if "_features" not in self.__dict__:
            raise AttributeError("Should be fitted at first.")

        feats = [
            "{0}__{1}".format(self._fname_prefix, x) if self._fname_prefix is not None else x for x in self._features
        ]

        return feats

    @features.setter
    def features(self, val: Sequence[str]):
        """Write input feature names.

        Args:
            val: Sequence of input features names.

        """
        self._features = deepcopy(val)

    def fit(self, dataset: LAMLDataset) -> "LAMLTransformer":
        """Fit transformer and return it's instance.

        Args:
            dataset: Dataset to fit on.

        Returns:
            self.

        """
        self.features = dataset.features
        for check_func in self._fit_checks:
            check_func(dataset)
        return self

    def transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Transform on dataset.

        Args:
            dataset: Dataset to make transform.

        Returns:
            LAMLDataset with new features.

        """
        for check_func in self._transform_checks:
            check_func(dataset)
        return dataset

    def fit_transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Default implementation of fit_transform - fit and then transform.

        Args:
            dataset: Dataset to fit and then transform on it.

        Returns:
            Dataset with new features.

        """
        for check_func in self._fit_checks:
            check_func(dataset)
        self.fit(dataset)

        for check_func in self._transform_checks:
            check_func(dataset)

        return self.transform(dataset)


class SequentialTransformer(LAMLTransformer):
    """Transformer that contains the list of transformers and apply one by one sequentially.

    Args:
        transformer_list: Sequence of transformers.


    """

    def __init__(self, transformer_list: Sequence[LAMLTransformer]):
        self.transformer_list = transformer_list

    def fit(self, dataset: LAMLDataset):
        """Fit not supported. Needs output to fit next transformer.

        Args:
            dataset: Dataset to fit.

        """
        raise NotImplementedError("Sequential supports only fit_transform.")

    def transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Apply the sequence of transformers to dataset one over output of previous.

        Args:
            dataset: Dataset to transform.

        Returns:
            Dataset with new features.

        """
        for trf in self.transformer_list:
            dataset = trf.transform(dataset)

        return dataset

    def fit_transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Sequential ``.fit_transform``.

         Output features - features from last transformer with no prefix.

        Args:
            dataset: Dataset to transform.

        Returns:
            Dataset with new features.

        """
        for trf in self.transformer_list:
            dataset = trf.fit_transform(dataset)

        self.features = self.transformer_list[-1].features

        return dataset


class UnionTransformer(LAMLTransformer):
    """Transformer that apply the sequence on transformers in parallel on dataset and concatenate the result.

    Args:
        transformer_list: Sequence of transformers.
        n_jobs: Number of processes to run fit and transform.

    """

    def __init__(self, transformer_list: Sequence[LAMLTransformer], n_jobs: int = 1):
        # TODO: Add multiprocessing version here
        self.transformer_list = [x for x in transformer_list if x is not None]
        self.n_jobs = n_jobs

    def _fit_singleproc(self, dataset: LAMLDataset) -> "UnionTransformer":
        """Single process version of fit.

        Args:
            dataset: Dataset to fit on.

        Returns:
            self.

        """
        # TODO: just for structure. Add parallel version later
        fnames = []
        for trf in self.transformer_list:
            trf.fit(dataset)
            fnames.append(trf.features)

        self.features = fnames

        return self

    def _fit_multiproc(self, dataset: LAMLDataset) -> "UnionTransformer":
        """Multi-process version of fit.

        Args:
            dataset: Dataset to fit on.

        Returns:  # noqa: DAR202
            self.

        """
        raise NotImplementedError

    def fit(self, dataset: LAMLDataset) -> "UnionTransformer":
        """Fit transformers in parallel.

         Output names - concatenation of features names with no prefix.

        Args:
            dataset: Dataset to fit on.

        Returns:
            self.

        """
        if self.n_jobs == 1:
            return self._fit_singleproc(dataset)
        else:
            return self._fit_multiproc(dataset)

    def _fit_transform_singleproc(self, dataset: LAMLDataset) -> List[LAMLDataset]:
        """Single process version of ``fit_transform``.

        Args:
            dataset: Dataset to transform.

        Returns:
            List of datasets with new features.

        """
        res = []
        fnames = []
        current = 0
        for n in range(len(self.transformer_list)):
            trf = self.transformer_list[n]
            ds = trf.fit_transform(dataset)
            if ds is not None:
                fnames.append(trf.features)
                res.append(ds)
                c = self.transformer_list[current]
                self.transformer_list[current] = trf
                self.transformer_list[n] = c
                current += 1
        self.transformer_list = self.transformer_list[:current]

        self.features = fnames

        return res

    def _fit_transform_multiproc(self, dataset: LAMLDataset) -> List[LAMLDataset]:
        """Multiproc version of fit_transform.

        Args:
            dataset: Dataset to fit on.

        Return:
            Now not implemented. Will be: Dataset with new features.

        """
        raise NotImplementedError

    def fit_transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Fit and transform transformers in parallel.

         Output names - concatenation of features names with no prefix.

        Args:
            dataset: Dataset to fit and transform on.

        Returns:
            Dataset with new features.

        """
        if self.n_jobs == 1:
            res = self._fit_transform_singleproc(dataset)
        else:
            res = self._fit_transform_multiproc(dataset)

        res = concatenate(res) if len(res) > 0 else None

        return res

    def _transform_singleproc(self, dataset: LAMLDataset) -> List[LAMLDataset]:
        """Single process version of transform.

        Args:
            dataset: Dataset to transform.

        Returns:
            List of dataset with new features.

        """
        res = []

        for trf in self.transformer_list:
            ds = trf.transform(dataset)
            res.append(ds)

        return res

    def _transform_multiproc(self, dataset: LAMLDataset) -> List[LAMLDataset]:
        """Multi-process version of transform.

        Args:
            dataset: Dataset to transform.

        Returns:  # noqa: DAR202
            List of datasets with new features.

        """
        raise NotImplementedError

    def transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Apply transformers in parallel.

         Output names - concatenation of features names with no prefix.

        Args:
            dataset: Dataset to fit and transform on.

        Returns:
            Dataset with new features.

        """
        if self.n_jobs == 1:
            res = self._transform_singleproc(dataset)
        else:
            res = self._transform_multiproc(dataset)

        res = concatenate(res)

        return res


class ColumnsSelector(LAMLTransformer):
    """Select columns to pass to another transformers (or feature selection).

    Args:
        keys: Columns names.

    """

    def __init__(self, keys: Sequence[str]):
        self.keys = keys

    def fit(self, dataset: LAMLDataset) -> "ColumnsSelector":
        """Empty fit method - just set features.

        Args:
            dataset: Dataset to fit on.

        Returns:
            self.

        """
        self.features = self.keys

        return self

    def transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Select given keys from dataset.

        Args:
            dataset: Dataset to transform.

        Returns:
            Dataset with selected features.

        """
        # to avoid coping if not needed
        if len(self.keys) == len(dataset.features) and all((x == y for (x, y) in zip(self.keys, dataset.features))):
            return dataset
        return dataset[:, self.keys]


class ColumnwiseUnion(UnionTransformer):
    # TODO: Union is not ABC !! NotImplemented - means not done right now
    """Apply 1 columns transformer to all columns.

    Example: encode all categories with single category encoders.

    Args:
        transformer: Dataset - base transformer.
        n_jobs: Number of threads.

    """

    def __init__(self, transformer: LAMLTransformer, n_jobs: int = 1):
        self.base_transformer = transformer
        self.n_jobs = n_jobs

    def _create_transformers(self, dataset: LAMLDataset):
        """Make a copies of base transformer.

        Args:
            dataset: Dataset with input features.

        """
        self.transformer_list = []

        for i in dataset.features:
            pipe = [ColumnsSelector([i]), deepcopy(self.base_transformer)]

            self.transformer_list.append(SequentialTransformer(pipe))

    def fit(self, dataset: LAMLDataset):
        """Create transformer list and then fit.

        Args:
            dataset: Dataset with input features.

        Returns:
            self.

        """
        self.features = dataset.features
        self._create_transformers(dataset)

        return super().fit(dataset)

    def fit_transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Create transformer list and then fit and transform.

        Args:
            dataset: Dataset with input features.

        Returns:
            Dataset with new features.

        """
        self.features = dataset.features
        self._create_transformers(dataset)

        return super().fit_transform(dataset)


class BestOfTransformers(LAMLTransformer):
    """Apply multiple transformers and select best."""

    def __init__(self, transformer_list: Sequence[LAMLTransformer], criterion: Callable):
        """Create selector from candidate list and selection criterion.

        Args:
            transformer_list: Sequence of transformers.
            criterion: Score function (greater is better).

        """
        self.transformer_list = transformer_list
        self.criterion = criterion

    def fit(self, dataset: LAMLDataset):
        """Empty method - raise error.

         This transformer supports only ``fit_transform``.

        Args:
            dataset: LAMLDataset to fit on.

        Raises:
            NotImplementedError: Always.

        """
        raise NotImplementedError("Support only fit_transform in BestOfTransformers")

    def fit_transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Fit transform all transformers and then select best.

        Args:
            dataset: Dataset to fit and then transform.

        Returns:
            Dataset with new features.

        """
        res = []

        for trf in self.transformer_list:
            ds = trf.fit_transform(dataset)
            res.append(ds)

        self.scores = np.array([self.criterion(ds) for ds in res])
        idx = self.scores.argmax()
        self.best_transformer = self.transformer_list[idx]
        self.features = self.best_transformer.features

        return res[idx]

    def transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Make transform by the best selected transformer.

        Args:
            dataset: Dataset with input features.

        Returns:
            Dataset with new features.

        """
        return self.best_transformer.transform(dataset)


class ConvertDataset(LAMLTransformer):
    """Convert dataset to given type.

    Args:
        dataset_type: Type to which to convert.

    """

    def __init__(self, dataset_type: Type[LAMLDataset]):
        self.dataset_type = dataset_type

    def transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Dataset type should implement ``from_dataset`` method.

        Args:
            dataset: Dataset to convert.

        Returns:
            Converted dataset.

        """
        return self.dataset_type.from_dataset(dataset)


class ChangeRoles(LAMLTransformer):
    """Change data roles (include dtypes etc).

    Args:
        roles: New roles for dataset.

    """

    def __init__(self, roles: Roles):
        self.roles = roles

    def transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Paste new roles into dataset.

        Args:
            dataset: Dataset to transform.

        Returns:
            New dataset.

        """
        data, features, _ = dataset.data, dataset.features, dataset.roles
        dataset = dataset.empty()
        dataset.set_data(data, features, self.roles)

        return dataset


class SetAttribute(LAMLTransformer):
    """Set data attribute.

    Args:
        attr: New attribute for dataset.
        column: Name of column.

    """

    def __init__(self, attr, column):

        self.attr = attr
        self.column = column

    def transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Paste new attributes into dataset.

        Args:
            dataset: Dataset to transform.

        Returns:
            Same dataset with new attributes.

        """
        val = dataset[:, self.column].data

        if isinstance(val, pd.DataFrame):
            val = val.iloc[:, 0]
        elif len(val.shape) == 2:
            val = val[:, 0]

        if self.attr not in dataset.__dict__:
            dataset.__dict__[self.attr] = val
            dataset._array_like_attrs.append(self.attr)

        return dataset


class EmptyTransformer(LAMLTransformer):
    """Set data attribute."""

    def __init__(self):
        pass

    def transform(self, dataset: LAMLDataset) -> LAMLDataset:
        """Paste new attributes into dataset.

        Args:
            dataset: Dataset to transform.

        Returns:
            Same dataset with new attributes.

        """
        print(dataset.__dict__)

        return dataset



"""Utils for new predict method in pytorch DataParallel."""

import threading

from itertools import chain
from typing import List
from typing import Optional

import torch
import torch.nn as nn

from torch._utils import ExceptionWrapper
from torch.cuda._utils import _get_device_index


def get_a_var(obj):  # noqa: D103
    if isinstance(obj, torch.Tensor):
        return obj

    if isinstance(obj, list) or isinstance(obj, tuple):
        for result in map(get_a_var, obj):
            if isinstance(result, torch.Tensor):
                return result
    if isinstance(obj, dict):
        for result in map(get_a_var, obj.items()):
            if isinstance(result, torch.Tensor):
                return result
    return None


def parallel_apply_predict(modules, inputs, kwargs_tup=None, devices=None):
    """Applies each `module` predict method in `modules` in parallel on arguments.

    contained in `inputs` (positional) and `kwargs_tup` (keyword)
    on each of `devices`.

    Args:
        modules: modules to be parallelized.
        inputs: inputs to the modules.
        kwargs_tup: Arguments for each modules.
        devices: CUDA devices.

    Returns:
        Predictions.

    """
    assert len(modules) == len(inputs)
    if kwargs_tup is not None:
        assert len(modules) == len(kwargs_tup)
    else:
        kwargs_tup = ({},) * len(modules)
    if devices is not None:
        assert len(modules) == len(devices)
    else:
        devices = [None] * len(modules)
    devices = list(map(lambda x: _get_device_index(x, True), devices))
    lock = threading.Lock()
    results = {}
    grad_enabled = torch.is_grad_enabled()

    def _worker(i, module, input, kwargs, device=None):
        torch.set_grad_enabled(grad_enabled)
        if device is None:
            device = get_a_var(input).get_device()
        try:
            with torch.cuda.device(device):
                # this also avoids accidental slicing of `input` if it is a Tensor
                if not isinstance(input, (list, tuple)):
                    input = (input,)
                output = module.predict(*input, **kwargs)
            with lock:
                results[i] = output
        except Exception:
            with lock:
                results[i] = ExceptionWrapper(where="in replica {} on device {}".format(i, device))

    if len(modules) > 1:
        threads = [
            threading.Thread(target=_worker, args=(i, module, input, kwargs, device))
            for i, (module, input, kwargs, device) in enumerate(zip(modules, inputs, kwargs_tup, devices))
        ]

        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
    else:
        _worker(0, modules[0], inputs[0], kwargs_tup[0], devices[0])

    outputs = []
    for i in range(len(inputs)):
        output = results[i]
        if isinstance(output, ExceptionWrapper):
            output.reraise()
        outputs.append(output)
    return outputs


class CustomDataParallel(nn.DataParallel):
    """Extension for nn.DataParallel for supporting predict method of DL model."""

    def __init__(
        self,
        module: nn.Module,
        device_ids: Optional[List[int]] = None,
        output_device: Optional[torch.device] = None,
        dim: Optional[int] = 0,
    ):
        super(CustomDataParallel, self).__init__(module, device_ids, output_device, dim)
        try:
            self.n_out = module.n_out
        except:
            pass

    def predict(self, *inputs, **kwargs):
        """Predict."""
        if not self.device_ids:
            return self.module(*inputs, **kwargs)

        for t in chain(self.module.parameters(), self.module.buffers()):
            if t.device != self.src_device_obj:
                raise RuntimeError(
                    "module must have its parameters and buffers "
                    "on device {} (device_ids[0]) but found one of "
                    "them on device: {}".format(self.src_device_obj, t.device)
                )

        inputs, kwargs = self.scatter(inputs, kwargs, self.device_ids)
        if len(self.device_ids) == 1:
            return self.module.predict(*inputs[0], **kwargs[0])
        replicas = self.replicate(self.module, self.device_ids[: len(inputs)])
        outputs = self.parallel_apply_predict(replicas, inputs, kwargs)
        return self.gather(outputs, self.output_device)

    def parallel_apply_predict(self, replicas, inputs, kwargs):
        """Parallel prediction."""
        return parallel_apply_predict(replicas, inputs, kwargs, self.device_ids[: len(replicas)])



"""Interpretation utils."""

import itertools

from collections import defaultdict
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import torch

from matplotlib.colors import Colormap
from torch.distributions.utils import clamp_probs


T_untokenized = Union[List[str], Tuple[List[str], List[Any]]]


class WrappedVocabulary:
    """Look-up table (word to token-id), handling unrecognized words.

    Args:
        word_to_id: Word to token id dictionary.
        unk_token: Token with which to replace unrecognized words.

    """

    def __init__(self, word_to_id: Dict[str, int], unk_token: str = "<UNK>"):
        self.word_to_id = word_to_id
        self.unk_token = word_to_id[unk_token]

    def __call__(self, x: str) -> int:
        """Map word to token-id."""
        return self.word_to_id.get(x, self.unk_token)

    def __getitem__(self, val: str) -> int:
        return self.word_to_id.get(val, self.unk_token)

    def __len__(self):
        return len(self.word_to_id)


class WrappedTokenizer:
    """Handler for automl tokenizer.

    Args:
        tokenizer: Automl tokenizer.

    """

    def __init__(self, tokenizer: "lightautoml.text.tokenizer.BaseTokenizer"):  # noqa F821
        self._tokenizer = tokenizer

    def __call__(self, x: str) -> List[str]:  # noqa D102
        return self._tokenizer.tokenize_sentence(self._tokenizer._tokenize(x))


def untokenize(
    raw: str,
    tokens: List[str],
    return_mask: bool = False,
    token_sym: Any = True,
    untoken_sym: Any = False,
) -> T_untokenized:
    """Get between tokens symbols.

    Args:
        raw: Raw string.
        tokens: List of tokens from raw string.
        return_mask: Flag to return mask
            for each new token. Format: list of
            `token_sym`, `untoken_sym`.
        token_sym: Object, denote token symbol.
        untoken_sym: Object, denote untoken symbol.

    Returns:
        Tuple (full_tokens, tokens_mask) if `return_mask=True`,
            else just list full_tokens.

    """
    mask = []
    untokenized = []
    pos = raw.find(tokens[0])

    if pos != 0:
        untokenized.append(raw[:pos])
        mask.append(untoken_sym)
        raw = raw[pos:]

    prev_token = tokens[0]
    for token in tokens[1:]:
        raw = raw[len(prev_token) :]
        pos = raw.find(token)
        untokenized.append(prev_token)
        mask.append(token_sym)
        if pos:
            mask.append(untoken_sym)
            untokenized.append(raw[:pos])
        prev_token = token
        raw = raw[pos:]

    untokenized.append(prev_token)
    mask.append(token_sym)

    cur = len(prev_token)
    if cur != len(raw):
        untokenized.append(raw[cur:])
        mask.append(untoken_sym)

    if return_mask:
        return untokenized, mask

    return untokenized


def find_positions(tokens: List[str], mask: List[bool]) -> List[int]:
    """Set positions and tokens.

    Args:
        tokens: List of tokens and untokens.
        mask: Mask for tokens.

    Returns:
        List of positions of tokens.

    """
    pos = []
    for i, (token, istoken) in enumerate(zip(tokens, mask)):
        if istoken:
            pos.append(i)

    return pos


class IndexedString:
    """Indexed string.

    Args:
        raw_string: Raw string.
        tokenizer: Tokenizer class.
        force_order: Save order, or use features as
            bag-of-words.

    """

    def __init__(self, raw_string: str, tokenizer: Any, force_order: bool = True):
        self.raw = raw_string
        self.tokenizer = tokenizer
        self.force_order = force_order

        self.toks_ = self._tokenize(raw_string)
        self.toks = [token.lower() for token in self.toks_]
        self.as_list_, self.mask = untokenize(self.raw, self.toks_, return_mask=True)

        self.pos = find_positions(self.as_list_, self.mask)
        self.as_np_ = np.array(self.as_list_)
        self.inv = []
        if not force_order:
            pos = defaultdict(list)
            self.vocab = {}
            for token, cur in zip(self.toks, self.pos):
                if token not in self.vocab:
                    self.vocab[token] = len(self.vocab)
                    self.inv.append(token)
                idx = self.vocab[token]
                pos[idx].append(cur)
            self.pos = pos
        else:
            self.inv = self.toks_

    def _tokenize(self, text: str) -> List[str]:
        prep_text = self.tokenizer._tokenize(text)
        tokens = self.tokenizer.tokenize_sentence(prep_text)

        return tokens

    def word(self, idx: int) -> str:
        """Token by its index.

        Args:
            idx: Index of token.

        Returns:
            Token.

        """
        return self.inv[idx]

    def inverse_removing(self, to_del: Union[List[str], List[int]], by_tokens: bool = False) -> str:
        """Remove tokens.

        Args:
            to_del: Tokens (text of int) to del.
            by_tokens: Flag if tokens are text or indexes.

        Returns:
            String without removed tokens.

        """
        # todo: this type of mapping will be not use order,
        # in case when we have not unique tokens.
        assert (not self.force_order) or (self.force_order and not by_tokens)

        if not self.force_order:
            if by_tokens:
                to_del = [self.t_i[token.lower()] for token in to_del]
                to_del = np.array(to_del)
            to_del = list(itertools.chain.from_iterable([self.pos[i] for i in to_del]))
        else:
            to_del = [self.pos[i] for i in to_del]
        mask = np.ones_like(self.as_np_, dtype=bool)
        mask[to_del] = False
        new_str = "".join(self.as_np_[mask])

        return new_str

    @property
    def n_words(self) -> int:
        """Number of unique words."""
        return len(self.pos)


def draw_html(
    tokens_and_weights: List[Tuple[str, float]],
    task_name: str,
    cmap: Any = None,
    grad_line: bool = True,
    grad_positive_label: Optional[str] = None,
    grad_negative_label: Optional[str] = None,
    prediction: Optional[float] = None,
    n_ticks: int = 10,
    draw_order: bool = False,
) -> str:
    """Get colored text in html format.

    For color used gradient from cmap.

    Args:
        tokens_and_weights: List of tokens.
        task_name: Task name.
        cmap: ```matplotlib.colors.Colormap``` or single color string like (#FFFFFF).
            By default blue-white-red linear gradient is used.
        grad_line: Grad line.
        grad_positive_label: Positive label text.
        grad_negative_label: Negative label text.
        prediction: Prediction.
        n_ticks: Number of ticks for plot.
        draw_order: Draw order.

    Returns:
        HTML like string.

    """
    font_style = "font-size:14px;"

    token_template = '<span style="background-color: {color_hex};">' "{token}" "</span>"

    ticks_template = (
        '<div style="border-left: 1px solid black; height: 18px; float: left; width: {}%;">'
        '<span style="position: relative; top: 1em; left: -{}em">'
        "{:.1f}"
        "</span>"
        "</div>"
    )

    gradient_full = "margin-left: 10%; " "margin-right: 10%; "

    gradient_styling = (
        "background: linear-gradient(90deg, {} 0%, {} 50%, {} 100%); "
        "border: 1px solid black; "
        "margin-left: {}em; "
        "margin-right: {}em; "
        "float: center; "
    )
    ticks_styling = "margin-left: {}em; " "margin-right: {}em; "
    norm_const = max(map(lambda x: abs(x[1]), tokens_and_weights))
    order = int("{:.2e}".format(norm_const).split("e")[1])
    order_s = "✕ {:.0e}".format(10 ** order)
    scale_word = "Scale"
    if not draw_order:
        order_s = ""
        scale_word = ""
    lord = 0.5 * len(order_s) + 1.5  # lenght order
    inorm_const = 1 / norm_const
    if cmap is None:
        cmap = plt.get_cmap("bwr")

    def get_color_hex(weight):
        if isinstance(cmap, Colormap):
            if task_name == "reg":
                rgba = cmap((-weight * inorm_const * 0.25 + 0.5), bytes=True)
            else:
                rgba = cmap((-weight * inorm_const * 0.25 + 0.5), bytes=True)
                # may be sigmoid for classifications, but hard to understand
                # rgba = cmap(1.0 / (1 + np.exp(weight * inorm_const)), bytes=True)
            return "#%02X%02X%02X" % rgba[:3]
        elif isinstance(cmap, str):
            if weight == 0.0:
                return "#FFFFFF"
            return cmap

    if prediction is None:
        prediction = ""
        pred_field = ""
    else:
        if task_name == "reg":
            prediction = "{:.1e}".format(prediction)
        else:
            prediction = "{:.3f}".format(prediction)
        pred_field = "AutoML's prediction"

    if grad_line:
        if task_name == "reg":
            grad_positive_label = "Positive"
            grad_negative_label = "Negative"
        elif task_name == "multiclass":
            grad_positive_label = "Class: " + grad_positive_label
            grad_negative_label = "Other classes"
        elif task_name == "binary":
            grad_positive_label = "Class: " + grad_positive_label
            grad_negative_label = "Class: " + grad_negative_label
    else:
        grad_positive_label = ""
        grad_negative_label = ""

    tokens_html = [
        token_template.format(token=token, color_hex=get_color_hex(weight)) for token, weight in tokens_and_weights
    ]

    if grad_line:
        between_ticks = [(100 / (n_ticks)) - 5e-2 * 6 / n_ticks if i <= n_ticks - 1 else 0 for i in range(n_ticks + 1)]
        ticks = np.linspace(-norm_const, norm_const, n_ticks + 1) / (10 ** (order))
        ticks_chart = " ".join(
            [ticks_template.format(t, 0.7 + 0.385 * (k < 0), k) for t, k in zip(between_ticks, ticks)]
        )
        grad_statement = """
        <p style="text-align: center">
            Class mapping
        </p>
        <div style="{}">
            <div id="grad" style="{}">
                <p style="text-align:left; margin-left: 1%; margin-right: 1%; color: white;">
                    {}
                    <span style="float:right;">
                        {}
                    </span>
                </p>
            </div>

            <div style="{}">
                {}
            </div>

            <div style="float: right; right: 0.75em; top: -3em; position: relative; font-weight: bold;">{}</div>
            <div style="float: right; right: -2em; top: -2.9em; position: relative; font-weight: bold;">{}</div>

            <div style="float: left; left: -5.5em; top: -4.42em; position: relative;  font-weight: bold;">{}</div>
            <div style="float: left; left: -8.22em; top: -2.9em; position: relative;  font-weight: bold;">{}</div>
        </div>
        """.format(
            gradient_full,
            gradient_styling,
            grad_negative_label,
            grad_positive_label,
            ticks_styling.format(lord, lord),
            ticks_chart,
            scale_word,
            order_s,
            pred_field,
            prediction,
        ).format(
            get_color_hex(-norm_const),
            get_color_hex(0.0),
            get_color_hex(norm_const),
            lord,
            lord,
        )
    else:
        grad_statement = ""

    raw_html = """
    <div>
        <p style="text-align: center">
            Text
        </p>
        <div style="border: 1px solid black;">
            <p style="{}; margin-left: 1%; margin-right: 1%;">{}</p>
        </div>
        {}
    </div>
    """.format(
        font_style, " ".join(tokens_html), grad_statement
    )

    return raw_html


def cross_entropy_multiple_class(input: torch.FloatTensor, target: torch.FloatTensor) -> torch.Tensor:
    """Cross entropy evaluation."""
    return torch.mean(torch.sum(-target * torch.log(clamp_probs(input)), dim=1))



"""Roles guess."""

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

import numpy as np

from joblib import Parallel
from joblib import delayed
from pandas import DataFrame
from pandas import Series

from ..dataset.np_pd_dataset import NumpyDataset
from ..dataset.np_pd_dataset import PandasDataset
from ..dataset.roles import CategoryRole
from ..dataset.roles import ColumnRole
from ..dataset.roles import NumericRole
from ..reader.utils import set_sklearn_folds
from ..transformers.base import ChangeRoles
from ..transformers.base import LAMLTransformer
from ..transformers.base import SequentialTransformer
from ..transformers.categorical import FreqEncoder
from ..transformers.categorical import LabelEncoder
from ..transformers.categorical import MultiClassTargetEncoder
from ..transformers.categorical import MultioutputTargetEncoder
from ..transformers.categorical import OrdinalEncoder
from ..transformers.categorical import TargetEncoder
from ..transformers.numeric import QuantileBinning


NumpyOrPandas = Union[NumpyDataset, PandasDataset]
RolesDict = Dict[str, ColumnRole]
Encoder = Union[TargetEncoder, MultiClassTargetEncoder]


def ginic(actual: np.ndarray, pred: np.ndarray) -> float:
    """Denormalized gini calculation.

    Args:
        actual: True values.
        pred: Predicted values.

    Returns:
        Metric value.

    """
    actual = np.asarray(actual)
    n = len(actual)
    a_s = actual[np.argsort(pred)]
    a_c = a_s.cumsum()

    if a_s.sum() == 0:
        return 0

    gini_sum = a_c.sum() / a_s.sum() - (n + 1) / 2.0
    return gini_sum / n


def gini_normalizedc(a: np.ndarray, p: np.ndarray) -> float:
    """Calculated normalized gini.

    Args:
        a: True values.
        p: Predicted values.

    Returns:
        Metric value.

    """
    ginic_aa = ginic(a, a)

    if ginic_aa:
        return ginic(a, p) / ginic_aa

    return 0


def gini_normalized(y_true: np.ndarray, y_pred: np.ndarray, empty_slice: Optional[np.ndarray] = None):
    """Calculate normalized gini index.

    Args:
        y_true: True values.
        y_pred: Predicted values.
        empty_slice: Mask.

    Returns:
        Gini value.

    """
    # TODO: CHECK ABOUT ZERO TARGET SUM
    if empty_slice is None:
        empty_slice = np.isnan(y_pred)
    elif empty_slice.all():
        return 0.0

    if y_pred.ndim == 1:
        y_pred = y_pred[:, np.newaxis]

    if empty_slice.ndim > 1:
        empty_slice = empty_slice[:, 0]

    sl = ~empty_slice

    if y_true.ndim == 1:
        y_true = y_true[:, np.newaxis]

    assert (
        y_pred.shape[1] == 1 or y_true.shape[1] == y_pred.shape[1]
    ), "Shape mismatch. Only calculate NxM vs NxM or Nx1 vs NxM"
    outp_size = y_true.shape[1]

    ginis = np.zeros((outp_size,), dtype=np.float32)

    for i in range(outp_size):
        j = min(i, y_pred.shape[1] - 1)

        yt = y_true[:, i][sl]
        yp = y_pred[:, j][sl]
        ginis[i] = gini_normalizedc(yt, yp)

    return np.abs(ginis).mean()


def get_target_and_encoder(train: NumpyOrPandas) -> Tuple[Any, type]:
    """Get target encoder and target based on dataset.

    Args:
        train: Dataset.

    Returns:
        (Target values, Target encoder).

    """
    train = train.empty().to_numpy()
    target = train.target

    if train.task.name == "multiclass":
        n_out = np.max(target) + 1
        target = target[:, np.newaxis] == np.arange(n_out)[np.newaxis, :]
        target = cast(np.ndarray, target).astype(np.float32)
        encoder = MultiClassTargetEncoder

    elif (train.task.name == "multi:reg") or (train.task.name == "multilabel"):
        target = cast(np.ndarray, target).astype(np.float32)
        encoder = MultioutputTargetEncoder
    else:
        encoder = TargetEncoder

    return target, encoder


def calc_ginis(data: np.ndarray, target: np.ndarray, empty_slice: Optional[np.ndarray] = None):
    """Calculate ginis for array of preditions.

    Args:
        data: np.ndarray.
        target: np.ndarray.
        empty_slice: np.ndarray.

    Returns:
        gini.

    """
    scores = np.zeros(data.shape[1])
    for n in range(data.shape[1]):
        sl = None
        if empty_slice is not None:
            sl = empty_slice[:, n]

        scores[n] = gini_normalized(target, data[:, n], empty_slice=sl)

    return scores


def _get_score_from_pipe(
    train,
    target,
    pipe: Optional[LAMLTransformer] = None,
    empty_slice: Optional[np.ndarray] = None,
) -> np.ndarray:
    """Get normalized gini index from pipeline.

    Args:
        train:  np.ndarray.
        target: np.ndarray.
        pipe: LAMLTransformer.
        empty_slice: np.ndarray.

    Returns:
        np.ndarray.

    """
    shape = train.shape

    if pipe is not None:
        train = pipe.fit_transform(train)

    data = train.data.reshape(shape + (-1,))
    scores = calc_ginis(data, target, empty_slice)

    return scores


def get_score_from_pipe(
    train: NumpyOrPandas,
    target: np.ndarray,
    pipe: Optional[LAMLTransformer] = None,
    empty_slice: Optional[np.ndarray] = None,
    n_jobs: int = 1,
) -> np.ndarray:
    """Get normalized gini index from pipeline.

    Args:
        train: np.ndarray.
        target: np.ndarray.
        pipe: LAMLTransformer.
        empty_slice: np.ndarray.
        n_jobs: int.

    Returns:
        np.ndarray.

    """
    shape = train.shape
    if n_jobs == 1:
        return _get_score_from_pipe(train, target, pipe, empty_slice)

    idx = np.array_split(np.arange(shape[1]), n_jobs)
    idx = [x for x in idx if len(x) > 0]
    n_jobs = len(idx)
    names = [[train.features[x] for x in y] for y in idx]

    if empty_slice is None:
        empty_slice = [None] * n_jobs
    else:
        empty_slice = [empty_slice[:, x] for x in idx]

    with Parallel(n_jobs=n_jobs, prefer="processes", backend="loky", max_nbytes=None) as p:
        res = p(
            delayed(_get_score_from_pipe)(train[:, name], target, pipe, sl) for (name, sl) in zip(names, empty_slice)
        )
    return np.concatenate(list(map(np.array, res)))


def get_numeric_roles_stat(
    train: NumpyOrPandas,
    subsample: Optional[Union[float, int]] = 100000,
    random_state: int = 42,
    manual_roles: Optional[RolesDict] = None,
    n_jobs: int = 1,
) -> DataFrame:
    """Calculate statistics about different encodings performances.

    We need it to calculate rules about advanced roles guessing.
    Only for numeric data.

    Args:
        train: Dataset.
        subsample: size of subsample.
        random_state: int.
        manual_roles: Dict.
        n_jobs: int.

    Returns:
        DataFrame.

    """
    if manual_roles is None:
        manual_roles = {}

    roles_to_identify = []
    flg_manual_set = []
    # check for train dtypes
    for f in train.features:
        role = train.roles[f]
        if role.name == "Numeric":
            roles_to_identify.append(f)
            flg_manual_set.append(f in manual_roles)

    res = DataFrame(
        columns=[
            "flg_manual",
            "unique",
            "unique_rate",
            "top_freq_values",
            "raw_scores",
            "binned_scores",
            "encoded_scores",
            "freq_scores",
            "nan_rate",
        ],
        index=roles_to_identify,
    )
    res["flg_manual"] = flg_manual_set

    if len(roles_to_identify) == 0:
        return res

    train = train[:, roles_to_identify].to_numpy()

    if train.folds is None:
        train.folds = set_sklearn_folds(train.task, train.target, cv=5, random_state=42, group=train.group)

    if subsample is not None:
        idx = np.random.RandomState(random_state + 1).permutation(train.shape[0])[:subsample]
        train = train[idx]

    data, target = train.data, train.target

    # check task specific
    target, encoder = get_target_and_encoder(train)

    # s3d = data.shape + (-1,)
    empty_slice = np.isnan(data)

    # check scores as is
    res["raw_scores"] = get_score_from_pipe(train, target, empty_slice=empty_slice, n_jobs=n_jobs)

    # check unique values
    unique_values = [np.unique(data[:, x][~np.isnan(data[:, x])], return_counts=True) for x in range(data.shape[1])]
    top_freq_values = np.array([max(x[1]) for x in unique_values])
    unique_values = np.array([len(x[0]) for x in unique_values])
    res["unique"] = unique_values
    res["top_freq_values"] = top_freq_values
    res["unique_rate"] = res["unique"] / train.shape[0]

    # check binned categorical score
    trf = SequentialTransformer([QuantileBinning(), encoder()])
    res["binned_scores"] = get_score_from_pipe(train, target, pipe=trf, empty_slice=empty_slice, n_jobs=n_jobs)

    # check label encoded scores
    trf = SequentialTransformer([ChangeRoles(CategoryRole(np.float32)), LabelEncoder(), encoder()])
    res["encoded_scores"] = get_score_from_pipe(train, target, pipe=trf, empty_slice=empty_slice, n_jobs=n_jobs)

    # check frequency encoding
    trf = SequentialTransformer([ChangeRoles(CategoryRole(np.float32)), FreqEncoder()])
    res["freq_scores"] = get_score_from_pipe(train, target, pipe=trf, empty_slice=empty_slice, n_jobs=n_jobs)

    res["nan_rate"] = empty_slice.mean(axis=0)

    return res


def calc_encoding_rules(
    stat: DataFrame,
    numeric_unique_rate: float = 0.999,
    max_to_3rd_rate: float = 1.1,
    binning_enc_rate: float = 2,
    raw_decr_rate: float = 1.1,
    max_score_rate: float = 0.2,
    abs_score_val: float = 0.04,
) -> DataFrame:
    """Calculate rules based on encoding stats.

    Args:
        stat: DataFrame
        numeric_unique_rate: float.
        max_to_3rd_rate: float.
        binning_enc_rate: float.
        raw_decr_rate: float.
        max_score_rate: float.
        abs_score_val: float.

    Returns:
        DataFrame.

    """
    scores_stat = stat[["raw_scores", "binned_scores", "encoded_scores", "freq_scores"]].values

    top_encodings = scores_stat.argsort(axis=1)[:, ::-1]
    sorted_scores = np.take_along_axis(scores_stat, top_encodings, axis=1)

    stat["max_to_3rd_rate"] = sorted_scores[:, 0] / sorted_scores[:, 2]
    stat["max_to_2rd_rate"] = sorted_scores[:, 0] / sorted_scores[:, 1]
    stat["max_score"] = scores_stat.max(axis=1)
    stat["max_score_rate"] = stat["max_score"] / stat["max_score"].max()

    # exact numbers are (my guess)
    # 1 - best score is raw score
    # 2 - top 1 score is binned, top 2 score is raw
    # 3 - 2 unique values
    # 4 - to many unique values
    # 5 - encoding type have no impact
    # 6 - binning encode wins with high rate
    # 7 - raw encoding looses to top with very small rate
    # 8 - feature is weak (lower than abs co and lower than % from max??)
    # 9 - set manually

    # rules

    stat["rule_0"] = top_encodings[:, 0] == 0
    stat["rule_1"] = top_encodings[:, :1].sum(axis=1) == 1
    stat["rule_2"] = stat["unique"] <= 2
    stat["rule_3"] = stat["unique_rate"] > numeric_unique_rate
    stat["rule_4"] = stat["max_to_3rd_rate"] < max_to_3rd_rate
    stat["rule_5"] = (top_encodings[:, 0] == 1) & (stat["max_to_3rd_rate"] > binning_enc_rate)
    stat["rule_6"] = (top_encodings[:, 1] == 0) & (stat["max_to_2rd_rate"] < raw_decr_rate)
    stat["rule_7"] = (stat["max_score_rate"] < max_score_rate) | (stat["max_score"] < abs_score_val)
    stat["rule_8"] = stat["flg_manual"]

    return stat


def rule_based_roles_guess(stat: DataFrame) -> Dict[str, ColumnRole]:
    """Create roles dict based on stats.

    Args:
        stat: DataFrame.

    Returns:
        Dict.

    """
    numbers = stat[stat[[x for x in stat.columns if "rule_" in x]].any(axis=1)].copy()
    categories = stat.drop(numbers.index)

    # define encoding types
    roles_dict = {}

    # rules to determinate handling type
    numbers["discrete_rule"] = (~numbers["rule_7"]) & ((numbers["binned_scores"] / numbers["raw_scores"]) > 2)
    categories["int_rule"] = categories["unique"] < 10
    categories["freq_rule"] = (categories["freq_scores"] / categories["encoded_scores"]) > 1.3
    categories["ord_rule"] = categories["unique_rate"] > 0.01

    # numbers with discrete features
    role = NumericRole(np.float32, discretization=True)
    feats = numbers[numbers["discrete_rule"]].index
    roles_dict = {**roles_dict, **{x: role for x in feats}}

    # classic numbers
    role = NumericRole(np.float32)
    feats = numbers[~numbers["discrete_rule"]].index
    roles_dict = {**roles_dict, **{x: role for x in feats}}

    # low cardinal categories
    # role = CategoryRole(np.float32, encoding_type='int')
    feats = categories[categories["int_rule"]].index
    ordinal = categories["ord_rule"][categories["int_rule"]].values
    roles_dict = {
        **roles_dict,
        **{x: CategoryRole(np.float32, encoding_type="int", ordinal=y) for (x, y) in zip(feats, ordinal)},
    }

    # frequency encoded feats
    # role = CategoryRole(np.float32, encoding_type='freq')
    feats = categories[categories["freq_rule"]].index
    ordinal = categories["ord_rule"][categories["freq_rule"]].values
    roles_dict = {
        **roles_dict,
        **{x: CategoryRole(np.float32, encoding_type="freq", ordinal=y) for (x, y) in zip(feats, ordinal)},
    }

    # categories left
    # role = CategoryRole(np.float32)
    feats = categories[(~categories["freq_rule"]) & (~categories["int_rule"])].index
    ordinal = categories["ord_rule"][(~categories["freq_rule"]) & (~categories["int_rule"])].values
    roles_dict = {
        **roles_dict,
        **{x: CategoryRole(np.float32, encoding_type="auto", ordinal=y) for (x, y) in zip(feats, ordinal)},
    }

    return roles_dict


def get_category_roles_stat(
    train: NumpyOrPandas,
    subsample: Optional[Union[float, int]] = 100000,
    random_state: int = 42,
    n_jobs: int = 1,
):
    """Search for optimal processing of categorical values.

    Categorical means defined by user or object types.

    Args:
        train: Dataset.
        subsample: size of subsample.
        random_state: seed of random numbers generator.
        n_jobs: number of jobs.

    Returns:
        result.

    """
    roles_to_identify = []

    dtypes = []

    # check for train dtypes
    for f in train.features:
        role = train.roles[f]
        if role.name == "Category" and role.encoding_type == "auto":
            roles_to_identify.append(f)
            dtypes.append(role.dtype)

    res = DataFrame(
        columns=[
            "unique",
            "top_freq_values",
            "dtype",
            "encoded_scores",
            "freq_scores",
            "ord_scores",
        ],
        index=roles_to_identify,
    )

    res["dtype"] = dtypes

    if len(roles_to_identify) == 0:
        return res

    train = train[:, roles_to_identify].to_pandas()

    if train.folds is None:
        train.folds = set_sklearn_folds(train.task, train.target.values, cv=5, random_state=42, group=train.group)

    if subsample is not None:
        idx = np.random.RandomState(random_state + 1).permutation(train.shape[0])[:subsample]
        train = train[idx]

    # check task specific
    target, encoder = get_target_and_encoder(train)

    empty_slice = train.data.isnull().values

    # check label encoded scores
    trf = SequentialTransformer([LabelEncoder(), encoder()])
    res["encoded_scores"] = get_score_from_pipe(train, target, pipe=trf, empty_slice=empty_slice, n_jobs=n_jobs)

    # check frequency encoding
    trf = FreqEncoder()
    res["freq_scores"] = get_score_from_pipe(train, target, pipe=trf, empty_slice=empty_slice, n_jobs=n_jobs)

    # check ordinal encoding
    trf = OrdinalEncoder()
    res["ord_scores"] = get_score_from_pipe(train, target, pipe=trf, empty_slice=empty_slice, n_jobs=n_jobs)

    return res


def calc_category_rules(
    stat: DataFrame,
) -> DataFrame:
    """Select best encoding for categories based on stats.

    Args:
        stat: DataFrame.

    Returns:
        DataFrame.

    """
    scores_stat = stat[["encoded_scores", "freq_scores", "ord_scores"]].values

    top_encodings = scores_stat.argsort(axis=1)[:, ::-1]
    stat["max_score"] = scores_stat.max(axis=1)

    # ordinal if
    # 1 - ordinal is top
    # 2 - 2 unique values
    stat["ord_rule_1"] = top_encodings[:, 0] == 2
    stat["ord_rule_2"] = stat["unique"] <= 2
    # freq encoding
    # 1 - freq is top and more than N unique values
    stat["freq_rule_1"] = top_encodings[:, 0] == 1

    # auto if
    # 1 - encoded is top
    stat["auto_rule_1"] = top_encodings[:, 0] == 0

    return stat


def rule_based_cat_handler_guess(stat: DataFrame) -> Dict[str, ColumnRole]:
    """Create roles dict based on stats.

    Args:
        stat: DataFrame.

    Returns:
        Dict.

    """
    # define encoding types
    roles_dict = {}

    # rules to determinate handling type
    freqs = stat[stat[[x for x in stat.columns if "freq_rule_" in x]].any(axis=1)]
    auto = stat[stat[[x for x in stat.columns if "auto_rule_" in x]].any(axis=1)]
    ordinals = stat[stat[[x for x in stat.columns if "ord_rule_" in x]].any(axis=1)]

    for enc_type, st in zip(["freq", "auto", "ord"], [freqs, auto, ordinals]):

        ordinal = False
        if enc_type == "ord":
            enc_type = "auto"
            ordinal = True

        feats = list(st.index)
        dtypes = list(st["dtype"])
        roles_dict = {
            **roles_dict,
            **{x: CategoryRole(dtype=d, encoding_type=enc_type, ordinal=ordinal) for x, d in zip(feats, dtypes)},
        }

    return roles_dict


def get_null_scores(
    train: NumpyOrPandas,
    feats: Optional[List[str]] = None,
    subsample: Optional[Union[float, int]] = 100000,
    random_state: int = 42,
) -> Series:
    """Get null scores.

    Args:
        train: Dataset
        feats: list of features.
        subsample: size of subsample.
        random_state: seed of random numbers generator.

    Returns:
        Series.

    """
    if feats is not None:
        train = train[:, feats].to_pandas()

    if subsample is not None:
        idx = np.random.RandomState(random_state + 1).permutation(train.shape[0])[:subsample]
        train = train[idx]

    # check task specific
    target, _ = get_target_and_encoder(train)

    empty_slice = train.data.isnull().values
    notnan = empty_slice.sum(axis=0)
    notnan = (notnan != train.shape[0]) & (notnan != 0)

    scores = np.zeros(train.shape[1])
    scores_ = calc_ginis(empty_slice[:, notnan], target, None)
    scores[notnan] = scores_

    res = Series(scores, index=train.features, name="max_score")

    return res



"""Utilities for working with the structure of a dataset."""

from typing import Callable
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from .base import LAMLDataset
from .np_pd_dataset import CSRSparseDataset
from .np_pd_dataset import NumpyDataset
from .np_pd_dataset import PandasDataset
from .roles import ColumnRole
from .seq_np_pd_dataset import SeqNumpyPandasDataset


# RoleType = TypeVar("RoleType", bound=ColumnRole)


def roles_parser(init_roles: Dict[Union[ColumnRole, str], Union[str, Sequence[str]]]) -> Dict[str, ColumnRole]:
    """Parser of roles.

    Parse roles from old format numeric:
    ``[var1, var2, ...]`` to ``{var1:numeric, var2:numeric, ...}``.

    Args:
        init_roles: Mapping between roles and feature names.

    Returns:
        Roles dict in format key - feature names, value - roles.

    """
    roles = {}
    for r in init_roles:

        feat = init_roles[r]

        if isinstance(feat, str):
            roles[feat] = r

        else:
            for f in init_roles[r]:
                roles[f] = r

    return roles


def get_common_concat(
    datasets: Sequence[LAMLDataset],
) -> Tuple[Callable, Optional[type]]:
    """Get concatenation function for datasets of different types.

    Takes multiple datasets as input and check,
    if is's ok to concatenate it and return function.

    Args:
        datasets: Sequence of datasets.

    Returns:
        Function, that is able to concatenate datasets.

    """
    # TODO: Add pandas + numpy via transforming to numpy?
    dataset_types = set([type(x) for x in datasets])

    # general - if single type, concatenation for that type
    if len(dataset_types) == 1:
        klass = list(dataset_types)[0]
        return klass.concat, None

    # np and sparse goes to sparse
    elif dataset_types == {NumpyDataset, CSRSparseDataset}:
        return CSRSparseDataset.concat, CSRSparseDataset

    elif dataset_types == {NumpyDataset, PandasDataset}:
        return numpy_and_pandas_concat, None

    elif (dataset_types == {NumpyDataset, SeqNumpyPandasDataset}) or (
        dataset_types == {PandasDataset, SeqNumpyPandasDataset}
    ):
        return numpy_or_pandas_and_seq_concat, None

    raise TypeError("Unable to concatenate dataset types {0}".format(list(dataset_types)))


def numpy_and_pandas_concat(datasets: Sequence[Union[NumpyDataset, PandasDataset]]) -> PandasDataset:
    """Concat of numpy and pandas dataset.

    Args:
        datasets: Sequence of datasets to concatenate.

    Returns:
        Concatenated dataset.

    """
    datasets = [x.to_pandas() for x in datasets]

    return PandasDataset.concat(datasets)


def numpy_or_pandas_and_seq_concat(
    datasets: Sequence[Union[NumpyDataset, PandasDataset, SeqNumpyPandasDataset]]
) -> Union[NumpyDataset, PandasDataset]:
    """Concat plain and sequential dataset.

    If both datasets have same size then concat them as plain, otherwise include seq dataset inside plain one.

    Args:
        datasets: one plain and one seq dataset.

    Returns:
        Concatenated dataset.

    """
    assert len(datasets) == 2, "should be 1 sequential and 1 plain dataset"
    # get 1 numpy / pandas dataset
    for n, dataset in enumerate(datasets):
        if isinstance(dataset, SeqNumpyPandasDataset):
            seq_dataset = dataset
        else:
            plain_dataset = dataset

    if len(seq_dataset.data) == len(plain_dataset):
        return SeqNumpyPandasDataset.concat([seq_dataset, plain_dataset.to_pandas()])
    else:
        if hasattr(plain_dataset, "seq_data"):
            plain_dataset.seq_data[seq_dataset.name] = seq_dataset
        else:
            plain_dataset.seq_data = {seq_dataset.name: seq_dataset}

        return plain_dataset


def concatenate(datasets: Sequence[LAMLDataset]) -> LAMLDataset:
    """Dataset concatenation function.

    Check if datasets have common concat function and then apply.
    Assume to take target/folds/weights etc from first one.

    Args:
        datasets: Sequence of datasets.

    Returns:
        Dataset with concatenated features.

    """
    conc, klass = get_common_concat([ds for ds in datasets if ds is not None])

    # this part is made to avoid setting first dataset of required type
    if klass is not None:

        n = 0
        for n, ds in enumerate(datasets):
            if type(ds) is klass:
                break

        datasets = [datasets[n]] + [x for (y, x) in enumerate(datasets) if n != y]

    return conc(datasets)



"""Tabular data utils."""

import os
import warnings

from copy import copy
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd

from joblib import Parallel
from joblib import delayed
from pandas import DataFrame
from sqlalchemy import create_engine, text


def get_filelen(fname: str) -> int:
    """Get length of csv file.

    Args:
        fname: File name.

    Returns:
        Length of file.

    """
    cnt_lines = -1
    with open(fname, "rb") as fin:
        for line in fin:
            if len(line.strip()) > 0:
                cnt_lines += 1
    return cnt_lines


def get_batch_ids(arr, batch_size):
    """Generator of batched sequences.

    Args:
        arr: Sequence.
        batch_size: Batch size.

    Yields:
        Sequential batches.

    """
    n = 0
    while n < len(arr):
        yield arr[n : n + batch_size]
        n += batch_size


def get_file_offsets(
    file: str, n_jobs: Optional[int] = None, batch_size: Optional[int] = None
) -> Tuple[List[int], List[int]]:
    """Get file offsets.

    Args:
        file: File path.
        n_jobs: Number of jobs for multiprocessing.
        batch_size: Batch size.

    Returns:
        Offsets tuple.

    """
    assert n_jobs is not None or batch_size is not None, "One of n_jobs or batch size should be defined"

    lens = []
    with open(file, "rb") as f:
        # skip header
        header_len = len(f.readline())
        # get row lens
        length = 0
        for row in f:
            if len(row.strip()) > 0:
                lens.append(length)
                length += len(row)

    lens = np.array(lens, dtype=np.int64) + header_len

    if batch_size:
        indexes = list(get_batch_ids(lens, batch_size))
    else:
        indexes = np.array_split(lens, n_jobs)

    offsets = [x[0] for x in indexes]
    cnts = [x.shape[0] for x in indexes]

    return offsets, cnts


def _check_csv_params(**read_csv_params: dict):
    """Validate parameters for func `read_csv`.

    Args:
        **read_csv_params: Read parameters.

    Returns:
        New parameters.

    """
    for par in ["skiprows", "nrows", "index_col", "header", "names", "chunksize"]:
        if par in read_csv_params:
            read_csv_params.pop(par)
            warnings.warn(
                "Parameter {0} will be ignored in parallel mode".format(par),
                UserWarning,
            )

    return read_csv_params


def read_csv_batch(file: str, offset, cnt, **read_csv_params):
    """Read batch of data from csv.

    Args:
        file: File path.
        offset: Start of file.
        cnt: Number of rows to read.
        **read_csv_params: Handler parameters.

    Returns:
        Read data.

    """
    read_csv_params = copy(read_csv_params)
    if read_csv_params is None:
        read_csv_params = {}

    try:
        usecols = read_csv_params.pop("usecols")
    except KeyError:
        usecols = None

    header = pd.read_csv(file, nrows=0, **read_csv_params).columns

    with open(file, "rb") as f:
        f.seek(offset)
        data = pd.read_csv(f, header=None, names=header, chunksize=None, nrows=cnt, usecols=usecols, **read_csv_params)

    return data


def read_csv(file: str, n_jobs: int = 1, **read_csv_params) -> DataFrame:
    """Read data from csv.

    Args:
        file: File path.
        n_jobs: Number of workers.
        **read_csv_params: Handler parameters.

    Returns:
        Read data.
    """
    if n_jobs == 1:
        return pd.read_csv(file, **read_csv_params)

    if n_jobs == -1:
        n_jobs = os.cpu_count()

    _check_csv_params(**read_csv_params)
    offsets, cnts = get_file_offsets(file, n_jobs)

    with Parallel(n_jobs) as p:
        res = p(
            delayed(read_csv_batch)(file, offset=offset, cnt=cnt, **read_csv_params)
            for (offset, cnt) in zip(offsets, cnts)
        )

    res = pd.concat(res, ignore_index=True)

    return res


class Batch:
    """Class to wraps batch of data in different formats.

    Default - batch of DataFrame.

    """

    @property
    def data(self) -> DataFrame:
        """Get data from Batch object.

        Returns:
            Data.

        """
        return self._data

    def __init__(self, data):
        self._data = data


class FileBatch(Batch):
    """Batch of csv file.

    Args:
        file: File path.
        offset: File start.
        cnt: Number of rows to read.
        read_csv_params: Additional params to :func:`pandas.read_csv`.

    """

    @property
    def data(self) -> DataFrame:
        """Get data from Batch object.

        Returns:
            Read data.

        """
        data_part = read_csv_batch(self.file, cnt=self.cnt, offset=self.offset, **self.read_csv_params)

        return data_part

    def __init__(self, file, offset, cnt, read_csv_params):
        self.file = file
        self.offset = offset
        self.cnt = cnt
        self.read_csv_params = read_csv_params


class BatchGenerator:
    """Abstract - generator of batches from data.

    Args:
        batch_size: Batch size. Default is ``None``, split by `n_jobs`.
        n_jobs: Number of processes to handle.

    """

    def __init__(self, batch_size, n_jobs):

        if n_jobs == -1:
            n_jobs = os.cpu_count()

        self.n_jobs = n_jobs
        self.batch_size = batch_size

    def __getitem__(self, idx) -> Batch:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError


class DfBatchGenerator(BatchGenerator):
    """Batch generator from :class:`~pandas.DataFrame`.

    Args:
        data: Data used for generator.
        n_jobs: Number of processes to handle.
        batch_size: Batch size. Default is ``None``, split by `n_jobs`.

    """

    def __init__(self, data: DataFrame, n_jobs: int = 1, batch_size: Optional[int] = None):
        super().__init__(batch_size, n_jobs)

        self.data = data

        if self.batch_size is not None:
            self.idxs = list(get_batch_ids(np.arange(data.shape[0]), batch_size))
        else:
            self.idxs = [x for x in np.array_split(np.arange(data.shape[0]), n_jobs) if len(x) > 0]

    def __len__(self) -> int:

        if self.batch_size is not None:
            return int(np.ceil(self.data.shape[0] / self.batch_size))

        return int(self.n_jobs)

    def __getitem__(self, idx):

        return Batch(self.data.iloc[self.idxs[idx]])


class FileBatchGenerator(BatchGenerator):
    """Generator of batches from file.

    Args:
        file: File path.
        n_jobs: Number of processes to handle.
        batch_size: Batch size. Default is ``None``, split by `n_jobs`.
        read_csv_params: Params of reading csv file.
            Look for :func:`pandas.read_csv` params.

    """

    def __init__(
        self,
        file,
        n_jobs: int = 1,
        batch_size: Optional[int] = None,
        read_csv_params: dict = None,
    ):
        super().__init__(batch_size, n_jobs)

        self.file = file
        self.offsets, self.cnts = get_file_offsets(file, n_jobs, batch_size)

        if read_csv_params is None:
            read_csv_params = {}

        self.read_csv_params = read_csv_params

    def __len__(self) -> int:
        return len(self.cnts)

    def __getitem__(self, idx):
        return FileBatch(self.file, self.offsets[idx], self.cnts[idx], self.read_csv_params)


class SqlDataSource:
    """Data wrapper for SQL connection.

    Args:
        connection_string: database url; for reference see
            https://docs.sqlalchemy.org/en/13/core/engines.html#database-urls
        query: SQL query to obtain data from
        index: optional index column to be removed from the query result; can be None, str of List[str]

    """

    def __init__(
        self,
        connection_string: str,
        query: str,
        index: Optional[Union[str, List[str]]] = None,
    ):
        self.engine = create_engine(connection_string)
        self.query = text(query)
        self.index = index
        self._data = None

    @property
    def data(self):
        """Get data associated with the query as :class:`~pandas.DataFrame`.

        Returns:
            :class:`~pandas.DataFrame`
        """
        if self._data is None:
            with self.engine.begin() as conn:
                self._data = pd.read_sql(self.query, conn, index_col=self.index)
        return self._data

    def get_batch_generator(self, n_jobs: int = 1, batch_size: int = None):
        """Access data with batch generator.

        Args:
            n_jobs: Number of processes to read file.
            batch_size: Number of entries in one batch.

        Returns:
            DfBatchGenerator object
        """
        return DfBatchGenerator(self.data, n_jobs, batch_size)


ReadableToDf = Union[str, np.ndarray, DataFrame, Dict[str, np.ndarray], Batch]


def read_data(
    data: ReadableToDf,
    features_names: Optional[Sequence[str]] = None,
    n_jobs: int = 1,
    read_csv_params: Optional[dict] = None,
) -> Tuple[DataFrame, Optional[dict]]:
    """Get :class:`~pandas.DataFrame` from different data formats.

    Note:
        Supported now data formats:

            - Path to ``.csv``, ``.parquet``, ``.feather`` files.
            - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`.
              For example, ``{'data': X...}``. In this case,
              roles are optional, but `train_features`
              and `valid_features` required.
            - :class:`pandas.DataFrame`.

    Args:
        data: Readable to DataFrame data.
        features_names: Optional features names if ``numpy.ndarray``.
        n_jobs: Number of processes to read file.
        read_csv_params: Params to read csv file.

    Returns:
        Tuple with read data and new roles mapping.

    """
    if read_csv_params is None:
        read_csv_params = {}
    # case - new process
    if isinstance(data, Batch):
        return data.data, None

    if isinstance(data, DataFrame):
        return data, None
    # case - single array passed to inference
    if isinstance(data, np.ndarray):
        return DataFrame(data, columns=features_names), None

    # case - dict of array args passed
    if isinstance(data, dict):
        df = DataFrame(data["data"], columns=features_names)
        upd_roles = {}
        for k in data:
            if k != "data":
                name = "__{0}__".format(k.upper())
                assert name not in df.columns, "Not supported feature name {0}".format(name)
                df[name] = data[k]
                upd_roles[k] = name
        return df, upd_roles

    if isinstance(data, str):
        if data.endswith(".feather"):
            # TODO: check about feather columns arg
            data = pd.read_feather(data)
            if read_csv_params["usecols"] is not None:
                data = data[read_csv_params["usecols"]]
            return data, None

        if data.endswith(".parquet"):
            return pd.read_parquet(data, columns=read_csv_params["usecols"]), None

        else:
            return read_csv(data, n_jobs, **read_csv_params), None

    if isinstance(data, SqlDataSource):
        return data.data, None

    raise ValueError("Input data format is not supported")


def read_batch(
    data: ReadableToDf,
    features_names: Optional[Sequence[str]] = None,
    n_jobs: int = 1,
    batch_size: Optional[int] = None,
    read_csv_params: Optional[dict] = None,
) -> Iterable[BatchGenerator]:
    """Read data for inference by batches for simple tabular data.

    Note:
        Supported now data formats:

            - Path to ``.csv``, ``.parquet``, ``.feather`` files.
            - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`.
              For example, ``{'data': X...}``. In this case,
              roles are optional, but `train_features`
              and `valid_features` required.
            - :class:`pandas.DataFrame`.

    Args:
        data: Readable to DataFrame data.
        features_names: Optional features names if ``numpy.ndarray``.
        n_jobs: Number of processes to read file.
        batch_size: Batch size.
        read_csv_params: Params to read csv file.

    Returns:
        Generator of batches.

    """
    if read_csv_params is None:
        read_csv_params = {}

    if isinstance(data, DataFrame):
        return DfBatchGenerator(data, n_jobs=n_jobs, batch_size=batch_size)

    # case - single array passed to inference
    if isinstance(data, np.ndarray):
        return DfBatchGenerator(
            DataFrame(data, columns=features_names),
            n_jobs=n_jobs,
            batch_size=batch_size,
        )

    if isinstance(data, str):
        if not (data.endswith(".feather") or data.endswith(".parquet")):
            return FileBatchGenerator(
                data, n_jobs, batch_size, read_csv_params
            )  # read_csv(data, n_jobs, **read_csv_params)

        else:
            data, _ = read_data(data, features_names, n_jobs, read_csv_params)
            return DfBatchGenerator(data, n_jobs=n_jobs, batch_size=batch_size)

    if isinstance(data, SqlDataSource):
        return data.get_batch_generator(n_jobs, batch_size)

    raise ValueError("Data type not supported")



"""Wrapped Catboost for tabular datasets."""

import logging

from copy import copy
from typing import Callable
from typing import Dict
from typing import Tuple
from typing import Union

import catboost as cb
import numpy as np

from pandas import Series

from ..dataset.np_pd_dataset import CSRSparseDataset
from ..dataset.np_pd_dataset import NumpyDataset
from ..dataset.np_pd_dataset import PandasDataset
from ..pipelines.selection.base import ImportanceEstimator
from ..pipelines.utils import get_columns_by_role
from ..utils.logging import LoggerStream
from ..validation.base import TrainValidIterator
from .base import TabularMLAlgo
from .tuning.base import Choice
from .tuning.base import Uniform


logger = logging.getLogger(__name__)
TabularDataset = Union[NumpyDataset, CSRSparseDataset, PandasDataset]


class BoostCB(TabularMLAlgo, ImportanceEstimator):
    """Gradient boosting on decision trees from catboost library.

    All available parameters listed in CatBoost documentation:

        - https://catboost.ai/docs/concepts/python-reference_parameters-list.html#python-reference_parameters-list

    freeze_defaults:

        - ``True`` :  params may be rewritten depending on dataset.
        - ``False``:  params may be changed only manually or with tuning.

    ``timer``: :class:`~lightautoml.utils.timer.Timer` instance or ``None``.

    """

    _name: str = "CatBoost"

    _default_params = {
        "task_type": "CPU",
        "thread_count": 4,
        "random_seed": 42,
        "num_trees": 3000,
        "learning_rate": 0.03,
        "l2_leaf_reg": 1e-2,
        "bootstrap_type": "Bernoulli",
        # "bagging_temperature": 1,
        "grow_policy": "SymmetricTree",
        "max_depth": 5,
        "min_data_in_leaf": 1,
        "one_hot_max_size": 10,
        "fold_permutation_block": 1,
        "boosting_type": "Plain",
        "boost_from_average": True,
        "od_type": "Iter",
        "od_wait": 100,
        "max_bin": 32,
        "feature_border_type": "GreedyLogSum",
        "nan_mode": "Min",
        # "silent": False,
        "verbose": 100,
        "allow_writing_files": False,
    }

    def _infer_params(self) -> Tuple[dict, int, int, Callable, Callable]:
        """Infer all parameters.

        Returns:
            Tuple (params, num_trees, early_stopping_rounds, fobj, feval).

        """
        params = copy(self.params)
        early_stopping_rounds = params.pop("od_wait")
        num_trees = params.pop("num_trees")

        loss = self.task.losses["cb"]
        fobj = loss.fobj_name
        feval = loss.metric_name

        if fobj not in ["RMSE", "LogLoss", "CrossEntropy", "Quantile", "MAE", "MAPE"]:
            params.pop("boost_from_average")

        # TODO: what if parameter of metric occurs in list of parameters of loss -> intersection...

        return params, num_trees, early_stopping_rounds, fobj, feval

    def init_params_on_input(self, train_valid_iterator: TrainValidIterator) -> dict:
        """Get model parameters depending on input dataset parameters.

        Args:
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Parameters of model.

        """
        rows_num = len(train_valid_iterator.train)
        dataset = train_valid_iterator.train
        self.task = train_valid_iterator.train.task

        if train_valid_iterator.train.dataset_type != "CSRSparseDataset":
            self._nan_rate = train_valid_iterator.train.to_pandas().nan_rate()

        try:
            self._le_cat_features = getattr(self, "_le_cat_features")
        except AttributeError:
            self._le_cat_features = get_columns_by_role(dataset, "Category", label_encoded=True)

        try:
            self._text_features = getattr(self, "_text_features")
        except AttributeError:
            self._text_features = get_columns_by_role(dataset, "Text")

        suggested_params = copy(self.default_params)

        if self.freeze_defaults:
            return suggested_params

        init_lr, ntrees, es = 0.01, 2000, 100
        if self.task.name == "binary":
            # TODO: with an increase in the number of columns, the learning rate decreases
            if rows_num <= 6000:
                init_lr = 0.02
                ntrees = 500

            elif rows_num <= 20000:
                init_lr = 0.035
                ntrees = 5000

            elif rows_num <= 50000:
                init_lr = 0.03
                ntrees = 5000

            elif rows_num <= 60000:
                init_lr = 0.05
                ntrees = 2000

            elif rows_num <= 100000:
                init_lr = 0.045
                ntrees = 1500

            elif rows_num <= 150000:
                init_lr = 0.045
                ntrees = 3000

            elif rows_num <= 300000:
                init_lr = 0.045
                ntrees = 2000

            else:
                init_lr = 0.05
                ntrees = 3000

        elif (self.task.name == "multiclass") or (self.task.name == "multi:reg") or (self.task.name == "multilabel"):
            init_lr = 0.03
            ntrees = 4000

            if rows_num <= 100000:
                ntrees = 3000

            elif rows_num <= 50000:
                ntrees = 3000

            elif rows_num <= 10000:
                ntrees = 3000

        elif self.task.name == "reg":
            init_lr = 0.05
            ntrees = 2000
            es = 300

        suggested_params["learning_rate"] = init_lr
        suggested_params["num_trees"] = ntrees
        suggested_params["od_wait"] = es

        return suggested_params

    def _get_default_search_spaces(self, suggested_params: Dict, estimated_n_trials: int) -> Dict:
        """Sample hyperparameters from suggested.

        Args:
            suggested_params: Dict with parameters.
            estimated_n_trials: Maximum number of hyperparameter estimation.

        Returns:
            Dict with sampled hyperparameters.

        """
        optimization_search_space = {}

        try:
            nan_rate = getattr(self, "_nan_rate")
        except AttributeError:
            nan_rate = 0

        optimization_search_space["max_depth"] = Uniform(low=3, high=7, q=1)

        if nan_rate > 0:
            optimization_search_space["nan_mode"] = Choice(options=["Max", "Min"])

        if estimated_n_trials > 20:
            optimization_search_space["l2_leaf_reg"] = Uniform(low=1e-8, high=10.0, log=True)

            # optimization_search_space['bagging_temperature'] = trial.suggest_loguniform(
            #     name='bagging_temperature',
            #     low=0.01,
            #     high=10.0,
            # )

        if estimated_n_trials > 50:
            optimization_search_space["min_data_in_leaf"] = Uniform(low=1, high=20, q=1)

            # the only case when used this parameter is when categorical columns more than 0
            if len(self._le_cat_features) > 0:
                optimization_search_space["one_hot_max_size"] = Uniform(low=3, high=10, q=1)

        return optimization_search_space

    def _get_pool(self, dataset: TabularDataset):

        try:
            self._le_cat_features = getattr(self, "_le_cat_features")
        except AttributeError:
            self._le_cat_features = get_columns_by_role(dataset, "Category", label_encoded=True)
        self._le_cat_features = self._le_cat_features if self._le_cat_features else None

        try:
            self._text_features = getattr(self, "_text_features")
        except AttributeError:
            self._text_features = get_columns_by_role(dataset, "Text")
        self._text_features = self._text_features if self._text_features else None

        dataset = dataset.to_pandas()
        data = dataset.data
        dtypes = data.dtypes.to_dict()
        if self._le_cat_features:
            dtypes = {**dtypes, **{x: "int" for x in self._le_cat_features}}
        # for future
        if self._text_features:
            dtypes = {**dtypes, **{x: "int" for x in self._text_features}}

        data = data.astype(dtypes)
        if self._le_cat_features:
            # copy was made in prev astype
            data.astype({x: "category" for x in self._le_cat_features}, copy=False)

        if dataset.target is not None:
            target, weights = self.task.losses["cb"].fw_func(dataset.target, dataset.weights)
        else:
            target, weights = dataset.target, dataset.weights

        pool = cb.Pool(
            data,
            label=target,
            weight=weights,
            feature_names=dataset.features,
            cat_features=self._le_cat_features,
            text_features=self._text_features,
        )

        return pool

    def fit_predict_single_fold(self, train: TabularDataset, valid: TabularDataset) -> Tuple[cb.CatBoost, np.ndarray]:
        """Implements training and prediction on single fold.

        Args:
            train: Train Dataset.
            valid: Validation Dataset.

        Returns:
            Tuple (model, predicted_values).

        """
        params, num_trees, early_stopping_rounds, fobj, feval = self._infer_params()

        cb_train = self._get_pool(train)
        cb_valid = self._get_pool(valid)

        model = cb.CatBoost(
            {
                **params,
                **{
                    "num_trees": num_trees,
                    "objective": fobj,
                    "eval_metric": feval,
                    "od_wait": early_stopping_rounds,
                    "task_type": "CPU"
                    if self.task.name in ["multilabel", "multi:reg"]
                    else params["task_type"],  # warning in TabularAutoML?
                },
            }
        )

        model.fit(cb_train, eval_set=cb_valid, log_cout=LoggerStream(logger, verbose_eval=100))

        val_pred = self._predict(model, cb_valid, params)

        return model, val_pred

    def predict_single_fold(self, model: cb.CatBoost, dataset: TabularDataset) -> np.ndarray:
        """Predict of target values for dataset.

        Args:
            model: CatBoost object.
            dataset: Test dataset.

        Returns:
            Predicted target values.

        """
        params = self._infer_params()[0]
        cb_test = self._get_pool(dataset)

        pred = self._predict(model, cb_test, params)

        return pred

    def get_features_score(self) -> Series:
        """Computes feature importance.

        Computes as mean values of feature importance, provided by CatBoost (PredictionValuesChange), per all models.

        Returns:
            Series with feature importances.

        """
        # TODO: Not worked if used text features, there is a ticket in their github
        assert self.is_fitted, "Model must be fitted to compute importance."
        imp = 0
        for model in self.models:
            imp = imp + model.get_feature_importance(
                type="FeatureImportance",
                prettified=False,
                thread_count=self.params["thread_count"],
            )

        imp = imp / len(self.models)

        return Series(imp, index=self.features).sort_values(ascending=False)

    def fit(self, train_valid: TrainValidIterator):
        """Just to be compatible with :class:`~lightautoml.pipelines.selection.base.ImportanceEstimator`.

        Args:
            train_valid: Classic cv-iterator.

        """
        self.fit_predict(train_valid)

    def _predict(self, model: cb.CatBoost, pool: cb.Pool, params):
        pred = None
        if (self.task.name == "multiclass") or (self.task.name == "multilabel"):
            pred = model.predict(pool, prediction_type="Probability", thread_count=params["thread_count"])
        elif self.task.name == "binary":
            pred = model.predict(pool, prediction_type="Probability", thread_count=params["thread_count"])[..., 1]
        elif (self.task.name == "reg") or (self.task.name == "multi:reg"):
            pred = model.predict(pool, thread_count=params["thread_count"])

        pred = self.task.losses["cb"].bw_func(pred)

        return pred



"""Internal representation of dataset in numpy, pandas and csr formats."""

import warnings

from copy import copy
from copy import deepcopy
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from typing import Union

import numpy as np
import pandas as pd

from pandas import DataFrame
from pandas import Series
from scipy import sparse

from ..tasks.base import Task
from .base import IntIdx
from .base import LAMLColumn
from .base import LAMLDataset
from .base import RolesDict
from .base import valid_array_attributes
from .np_pd_dataset import CSRSparseDataset
from .np_pd_dataset import NumpyDataset
from .np_pd_dataset import PandasDataset
from .roles import ColumnRole


NpFeatures = Union[Sequence[str], str, None]
NpRoles = Union[Sequence[ColumnRole], ColumnRole, RolesDict, None]
DenseSparseArray = Union[np.ndarray, sparse.csr_matrix]
FrameOrSeries = Union[DataFrame, Series]
Dataset = TypeVar("Dataset", bound=LAMLDataset)
RowSlice = Optional[Union[Sequence[int], Sequence[bool]]]
ColSlice = Optional[Union[Sequence[str], str]]


class SeqNumpyPandasDataset(PandasDataset):
    """Sequential Dataset, that contains info in pd.DataFrame format."""

    _dataset_type = "SeqNumpyPandasDataset"

    def _check_dtype(self):
        """Check if dtype in .set_data is ok and cast if not."""
        date_columns = []

        self.dtypes = {}
        for f in self.roles:
            if self.roles[f].name == "Datetime":
                date_columns.append(f)
            else:
                self.dtypes[f] = self.roles[f].dtype

        self.data = self.data.astype(self.dtypes)
        self.data.reset_index(drop=True, inplace=True)
        # do we need to reset_index ?? If yes - drop for Series attrs too
        # case to check - concat pandas dataset and from numpy to pandas dataset
        # TODO: Think about reset_index here
        # self.data.reset_index(inplace=True, drop=True)

        # handle dates types
        for i in date_columns:
            dt_role = self.roles[i]
            if not (self.data.dtypes[i] is np.datetime64):
                self.data[i] = pd.to_datetime(
                    self.data[i], format=dt_role.format, unit=dt_role.unit, origin=dt_role.origin, cache=True
                )

            self.dtypes[i] = np.datetime64

    @property
    def idx(self) -> Any:
        """Get idx attribute.

        Returns:
            Any, array like or ``None``.

        """
        return self._idx

    @idx.setter
    def idx(self, val: Any):
        """Set idx array or ``None``.

        Args:
            val: Some idx or ``None``.

        """
        self._idx = val

    def _initialize(self, task: Optional[Task], **kwargs: Any):
        """Initialize empty dataset with task and array like attributes.

        Args:
            task: Task name for dataset.
            **kwargs: 1d arrays like attrs like target, group etc.

        """
        assert all([x in valid_array_attributes for x in kwargs]), "Unknown array attribute. Valid are {0}".format(
            valid_array_attributes
        )

        self.task = task
        # here we set target and group and so ...
        self._array_like_attrs = []
        for k in kwargs:
            self._array_like_attrs.append(k)
            self.__dict__[k] = kwargs[k]

        # checks for valid values in target, groups ...
        for check in self._init_checks:
            check(self)

        # set empty attributes
        self._idx = None
        self._data = None
        self._features = []
        self._roles = {}

    def __init__(
        self,
        data: Optional[DenseSparseArray],
        features: NpFeatures = (),
        roles: NpRoles = None,
        idx: List = (),
        task: Optional[Task] = None,
        name: Optional[str] = "seq",
        scheme: Optional[dict] = None,
        **kwargs: np.ndarray
    ):
        """Create dataset from `pd.DataFrame` and `pd.Series`.

        Args:
            data: Table with features.
            features: features names.
            roles: Roles specifier.
            idx: sequential indexes. Each element consists of corresponding sequence in data table.
            task: Task specifier.
            name: name of current dataset.
            scheme: dict of relations of current dataset with others.
            **kwargs: Series, array like attrs target, group etc...

        """
        self.name = name
        if scheme is not None:
            self.scheme = scheme
        else:
            self.scheme = {}

        self._initialize(task, **kwargs)
        if data is not None:
            self.set_data(data, roles, idx)

    def set_data(self, data: DenseSparseArray, roles: NpRoles = None, idx: Optional[List] = None):
        """Inplace set data, features, roles for empty dataset.

        Args:
            data: 2d array like or ``None``.
            roles: roles dict.
            idx: list.

        """
        # assert data is None or type(data) is np.ndarray, 'Numpy dataset support only np.ndarray features'
        super().set_data(data, None, roles)
        if idx is None:
            idx = np.arange(len(data)).reshape(-1, 1)
        self.idx = idx
        self._check_dtype()

    def __len__(self):
        return len(self.idx)

    def _get_cols_idx(self, columns: Union[Sequence[str], str]) -> Union[Sequence[int], int]:
        """Get numeric index of columns by column names.

        Args:
            columns: sequence of columns of single column.

        Returns:
            sequence of int indexes or single int.

        """
        if type(columns) is str:
            idx = self.data.columns.get_loc(columns)

        else:
            idx = self.data.columns.get_indexer(columns)

        return idx

    def __getitem__(self, k: Tuple[RowSlice, ColSlice]) -> Union["LAMLDataset", LAMLColumn]:
        """Select a subset of dataset.

        Define how to slice a dataset
        in way ``dataset[[1, 2, 3...], ['feat_0', 'feat_1'...]]``.
        Default behavior based on ``._get_cols``, ``._get_rows``, ``._get_2d``.

        Args:
            k: First element optional integer columns indexes,
                second - optional feature name or list of features names.

        Returns:
            Subset.

        """
        # TODO: Maybe refactor this part?
        if type(k) is tuple:
            rows, cols = k
            if isinstance(cols, str):
                cols = [cols]
        else:
            rows = k
            cols = None

        is_slice = False
        if isinstance(rows, slice):
            is_slice = True

        rows = [rows] if isinstance(rows, int) else np.arange(self.__len__()) if isinstance(rows, slice) else rows
        temp_idx = self.idx[rows]
        rows = []
        idx_new = []
        _c = 0
        for i in temp_idx:
            rows.extend(list(i))
            idx_new.append(list(np.arange(len(i)) + _c))
            _c += len(i)
        idx_new = np.array(idx_new, dtype=object)

        rows = np.array(sorted(list(set(rows))))

        if is_slice:
            idx_new = self.idx
            rows = np.arange(len(self.data))
        else:
            warnings.warn(
                "Resulted sequential dataset may have different structure. It's not recommended to slice new dataset"
            )

        # case when columns are defined
        if cols is not None:
            idx = self._get_cols_idx(cols)
            data = self._get_2d(self.data, (rows, idx))

            # case of multiple columns - return LAMLDataset
            roles = dict(((x, self.roles[x]) for x in self.roles if x in cols))
        else:
            roles = self.roles
            data = self._get_rows(self.data, rows)

        # case when rows are defined
        if rows is None:
            dataset = self.empty()
        else:
            dataset = copy(self)
            params = dict(((x, self._get_rows(self.__dict__[x], rows)) for x in self._array_like_attrs))
            dataset._initialize(self.task, **params)

        dataset.set_data(data, roles, idx=idx_new)

        return dataset

    def to_sequence(self, k: Tuple[RowSlice, ColSlice] = None) -> Union["LAMLDataset", LAMLColumn]:
        """Select a subset of dataset and transform it to sequence.

        Define how to slice a dataset in way ``dataset[[1, 2, 3...], ['feat_0', 'feat_1'...]]``.
        Default behavior based on ``._get_cols``, ``._get_rows``, ``._get_2d``.

        Args:
            k: First element optional integer columns indexes,
                second - optional feature name or list of features names.

        Returns:
            Numpy Dataset with new sequential dimension.

        """
        self._check_dtype()
        if k is None:
            k = slice(None, None, None)

        # TODO: Maybe refactor this part?
        if type(k) is tuple:
            rows, cols = k
            if isinstance(cols, str):
                cols = [cols]
        else:
            rows = k
            cols = None

        rows = [rows] if isinstance(rows, int) else np.arange(self.__len__()) if isinstance(rows, slice) else rows

        # case when columns are defined
        if cols is not None:
            idx = self._get_cols_idx(cols)

            # case when seqs have different shape, return array with arrays
            if len(self.idx.shape) == 1:
                data = []
                _d = self.data.iloc[:, idx].values
                for row in rows:
                    data.append(_d[self.idx[row]])
                data = np.array(data, dtype=object)
            else:
                data = self._get_3d(self.data, (self.idx[rows], idx))

            # case of multiple columns - return LAMLDataset
            roles = dict(((x, self.roles[x]) for x in self.roles if x in cols))
            features = [x for x in cols if x in set(self.features)]
        else:
            roles, features = self.roles, self.features

            if len(self.idx.shape) == 1:
                data = []
                _d = self.data.values
                for row in rows:
                    data.append(_d[self.idx[row]])
                data = np.array(data, dtype=object)
            else:
                data = self._get_3d(self.data, (self.idx[rows], self._get_cols_idx(self.data.columns)))

        # case when rows are defined
        if rows is None:
            dataset = NumpyDataset(None, features, deepcopy(roles), task=self.task)
        else:
            dataset = NumpyDataset(data, features, deepcopy(roles), task=self.task)

        return dataset

    def apply_func(self, k: Tuple[RowSlice, ColSlice] = None, func: Callable = None) -> np.ndarray:
        """Apply function to each sequence.

        Args:
            k: First element optional integer columns indexes,
                second - optional feature name or list of features names.
            func: any callable function

        Returns:
            output np.ndarray

        """
        self._check_dtype()
        if k is None:
            k = slice(None, None, None)

        # TODO: Maybe refactor this part?
        if type(k) is tuple:
            rows, cols = k
            if isinstance(cols, str):
                cols = [cols]
        else:
            rows = k
            cols = None

        rows = [rows] if isinstance(rows, int) else np.arange(self.__len__()) if isinstance(rows, slice) else rows

        # case when columns are defined
        if cols is not None:
            idx = self._get_cols_idx(cols)

            # case when seqs have different shape, return array with arrays
            data = []
            _d = self.data.iloc[:, idx].values
            for row in rows:
                data.append(func(_d[self.idx[row]]))
            data = np.array(data)
        else:
            data = []
            _d = self.data.values
            for row in rows:
                data.append(func(_d[self.idx[row]]))
            data = np.array(data)

        return data

    @classmethod
    def _get_2dT(cls, data: np.ndarray, k: Tuple[IntIdx, IntIdx]) -> np.ndarray:
        """Get 2d slice.

        Args:
            data: Data.
            k: Tuple of integer sequences.

        Returns:
            2d slice.

        """
        rows, cols = k

        return data[rows][:, cols]

    @classmethod
    def _get_3d(cls, data: np.ndarray, k: Tuple[IntIdx, IntIdx]) -> np.ndarray:
        """Get 3d slice.

        Args:
            data: Data.
            k: Tuple of integer sequences.

        Returns:
            3d slice.

        """
        rows, cols = k

        return data.iloc[:, cols].values[rows]

    def to_csr(self) -> "CSRSparseDataset":
        """Convert to csr.

        # noqa: DAR202
        Returns:
            Same dataset in CSRSparseDatatset format.

        """
        raise NotImplementedError

    def to_numpy(self) -> "NumpyDataset":
        """Convert to class:`NumpyDataset`.

        Returns:
            Same dataset in class:`NumpyDataset` format without sequential features.

        """
        # check for empty
        data = None if self.data is None else self.data.values
        roles = self.roles
        features = self.features
        # target and etc ..
        params = dict(((x, self.__dict__[x].values) for x in self._array_like_attrs))
        task = self.task

        return NumpyDataset(data, features, roles, task, **params)

    def to_pandas(self) -> "PandasDataset":
        """Convert to plain PandasDataset.

        Returns:
            Same dataset in PandasDataset format without sequential features.

        """
        # check for empty case
        data = None if self.data is None else DataFrame(self.data, columns=self.features)
        roles = self.roles
        # target and etc ..
        params = dict(((x, Series(self.__dict__[x])) for x in self._array_like_attrs))
        task = self.task

        return PandasDataset(data, roles, task, **params)

    @classmethod
    def concat(cls, datasets: Sequence["LAMLDataset"]) -> "LAMLDataset":
        """Concat multiple dataset.

        Default behavior - takes empty dataset from datasets[0]
        and concat all features from others.

        Args:
            datasets: Sequence of datasets.

        Returns:
            Concated dataset.

        """
        for check in cls._concat_checks:
            check(datasets)

        idx = datasets[0].idx
        dataset = datasets[0].empty()
        data = []
        features = []
        roles = {}

        atrs = set(dataset._array_like_attrs)
        for ds in datasets:
            data.append(ds.data)
            features.extend(ds.features)
            roles = {**roles, **ds.roles}
            for atr in ds._array_like_attrs:
                if atr not in atrs:
                    dataset._array_like_attrs.append(atr)
                    dataset.__dict__[atr] = ds.__dict__[atr]
                    atrs.update({atr})

        data = cls._hstack(data)
        dataset.set_data(data, roles, idx=idx)

        return dataset



"""HypEx Addon for LightAutoML.

This module forwards all imports from the official HypEx package,
maintaining the same API structure as in the original library.

Requirements:
    - Install LightAutoML with HypEx support:
      `pip install lightautoml[hypex]`

Examples:
    Importing models and utilities as in HypEx:

    >>> from lightautoml.addons.hypex import AATest
    >>> from lightautoml.addons.hypex.utils.tutorial_data_creation import create_test_data

    Creating test data:
    >>> some_large_dataframe = create_test_data(
    ...     rs=52, na_step=10, nan_cols=['age', 'gender'], num_users=100_000
    ... )

Raises:
    ImportError: If HypEx is not installed.
"""

import importlib
import sys

MODULE_NAME = "hypex"

try:
    hypex = importlib.import_module(MODULE_NAME)
except ImportError:
    raise ImportError(
        f"{MODULE_NAME} is not installed. Please install it using " f"'pip install lightautoml[{MODULE_NAME}]'."
    )

sys.modules["lightautoml.addons.hypex"] = hypex



"""Base AutoML class."""

import logging

from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence

from ..dataset.base import LAMLDataset
from ..dataset.utils import concatenate
from ..pipelines.ml.base import MLPipeline
from ..reader.base import Reader
from ..utils.logging import set_stdout_level
from ..utils.logging import verbosity_to_loglevel
from ..utils.timer import PipelineTimer
from ..validation.utils import create_validation_iterator
from .blend import BestModelSelector
from .blend import Blender
from joblib import dump


logger = logging.getLogger(__name__)


class AutoML:
    """Class for compile full pipeline of AutoML task.

    AutoML steps:

        - Read, analyze data and get inner
          :class:`~lightautoml.dataset.base.LAMLDataset` from input
          dataset: performed by reader.
        - Create validation scheme.
        - Compute passed ml pipelines from levels.
          Each element of levels is list
          of :class:`~lightautoml.pipelines.ml.base.MLPipelines`
          prediction from current level are passed to next level
          pipelines as features.
        - Time monitoring - check if we have enough time to calc new pipeline.
        - Blend last level models and prune useless pipelines
          to speedup inference: performed by blender.
        - Returns prediction on validation data.
          If crossvalidation scheme is used,
          out-of-fold prediction will returned.
          If validation data is passed
          it will return prediction on validation dataset.
          In case of cv scheme when some point of train data
          never was used as validation (ex. timeout exceeded
          or custom cv iterator like
          :class:`~lightautoml.validation.np_iterators.TimeSeriesIterator`
          was used) NaN for this point will be returned.

    Example:
        Common usecase - create custom pipelines or presets.

        >>> reader = SomeReader()
        >>> pipe = MLPipeline([SomeAlgo()])
        >>> levels = [[pipe]]
        >>> automl = AutoML(reader, levels, )
        >>> automl.fit_predict(data, roles={'target': 'TARGET'})

    Args:
        reader: Instance of Reader class object that
            creates :class:`~lightautoml.dataset.base.LAMLDataset`
            from input data.
        levels: List of list
            of :class:`~lightautoml.pipelines.ml..base.MLPipelines`.
        timer: Timer instance of
            :class:`~lightautoml.utils.timer.PipelineTimer`.
            Default - unlimited timer.
        blender: Instance of Blender.
            Default - :class:`~lightautoml.automl.blend.BestModelSelector`.
        skip_conn: True if we should pass first level
            input features to next levels.

    Note:
        There are several verbosity levels:

            - `0`: No messages.
            - `1`: Warnings.
            - `2`: Info.
            - `3`: Debug.

    """

    def __init__(
        self,
        reader: Reader,
        levels: Sequence[Sequence[MLPipeline]],
        timer: Optional[PipelineTimer] = None,
        blender: Optional[Blender] = None,
        skip_conn: bool = False,
        return_all_predictions: bool = False,
        debug: bool = False,
    ):
        self._initialize(reader, levels, timer, blender, skip_conn, return_all_predictions, debug)

    def _initialize(
        self,
        reader: Reader,
        levels: Sequence[Sequence[MLPipeline]],
        timer: Optional[PipelineTimer] = None,
        blender: Optional[Blender] = None,
        skip_conn: bool = False,
        return_all_predictions: bool = False,
        debug: bool = False,
    ):
        """Same as __init__. Exists for delayed initialization in presets.

        Args:
            reader: Instance of Reader class object that
                creates :class:`~lightautoml.dataset.base.LAMLDataset` from input data.
            levels: List of list of :class:`~lightautoml.pipelines.ml..base.MLPipelines`.
            timer: Timer instance of :class:`~lightautoml.utils.timer.PipelineTimer`.
                Default - unlimited timer.
            blender: Instance of Blender. Default - :class:`~lightautoml.automl.blend.BestModelSelector`.
            skip_conn: True if we should pass first level
                input features to next levels.
            return_all_predictions: True if we should return all predictions from last
                level models.
            debug: To catch running model exceptions or not.
                - ``True`` :  show exceptions during model training.
                - ``False``:  catch and hide exceptions.

        """
        assert len(levels) > 0, "At least 1 level should be defined"

        self.timer = timer
        if timer is None:
            self.timer = PipelineTimer()
        self.reader = reader
        self._levels = levels

        # default blender is - select best model and prune other pipes
        self.blender = blender
        if blender is None:
            self.blender = BestModelSelector()

        # update model names
        for i, lvl in enumerate(self._levels):

            for j, pipe in enumerate(lvl):
                pipe.upd_model_names("Lvl_{0}_Pipe_{1}".format(i, j))

        self.skip_conn = skip_conn
        self.return_all_predictions = return_all_predictions
        self.debug = debug

    def fit_predict(
        self,
        train_data: Any,
        roles: dict,
        train_features: Optional[Sequence[str]] = None,
        cv_iter: Optional[Iterable] = None,
        valid_data: Optional[Any] = None,
        valid_features: Optional[Sequence[str]] = None,
        verbose: int = 0,
        path_to_save: Optional[str] = None,
    ) -> LAMLDataset:
        """Fit on input data and make prediction on validation part.

        Args:
            train_data: Dataset to train.
            roles: Roles dict.
            train_features: Optional features names,
                if cannot be inferred from train_data.
            cv_iter: Custom cv iterator. For example,
                :class:`~lightautoml.validation.np_iterators.TimeSeriesIterator`.
            valid_data: Optional validation dataset.
            valid_features: Optional validation dataset
                features if can't be inferred from `valid_data`.
            verbose: Controls the verbosity: the higher, the more messages.
                <1  : messages are not displayed;
                >=1 : the computation process for layers is displayed;
                >=2 : the information about folds processing is also displayed;
                >=3 : the hyperparameters optimization process is also displayed;
                >=4 : the training process for every algorithm is displayed.
            path_to_save: The path that joblib will use to save the model after fit stage is completed. Use *.joblib format.

        Returns:
            Predicted values.

        """
        set_stdout_level(verbosity_to_loglevel(verbose))
        self.timer.start()
        train_dataset = self.reader.fit_read(train_data, train_features, roles)

        # Saving class mapping
        if self.reader.task.name == "binary":
            self.targets_order = [1]
        elif self.reader.task.name == "multi:reg":
            self.targets_order = roles["target"]
        elif self.reader.task.name == "reg":
            self.targets_order = [roles["target"]]
        elif self.reader.task.name == "multilabel":
            self.targets_order = roles["target"]
        else:  # multiclass
            self.targets_order = (
                sorted(self.reader.class_mapping, key=self.reader.class_mapping.get, reverse=False)
                if self.reader.class_mapping
                else None
            )

        assert (
            len(self._levels) <= 1 or train_dataset.folds is not None
        ), "Not possible to fit more than 1 level without cv folds"

        assert (
            len(self._levels) <= 1 or valid_data is None
        ), "Not possible to fit more than 1 level with holdout validation"

        valid_dataset = None
        if valid_data is not None:
            valid_dataset = self.reader.read(valid_data, valid_features, add_array_attrs=True)

        train_valid = create_validation_iterator(train_dataset, valid_dataset, n_folds=None, cv_iter=cv_iter)
        # for pycharm)
        level_predictions = None
        pipes = None

        self.levels = []

        for leven_number, level in enumerate(self._levels, 1):
            pipes = []
            level_predictions = []
            flg_last_level = leven_number == len(self._levels)

            logger.info(
                f"Layer \x1b[1m{leven_number}\x1b[0m train process start. Time left {self.timer.time_left:.2f} secs"
            )

            for k, ml_pipe in enumerate(level):
                ml_pipe.debug = self.debug
                pipe_pred = ml_pipe.fit_predict(train_valid)
                level_predictions.append(pipe_pred)
                pipes.append(ml_pipe)

                logger.info("Time left {:.2f} secs\n".format(self.timer.time_left))

                if self.timer.time_limit_exceeded():
                    logger.info(
                        "Time limit exceeded. Last level models will be blended and unused pipelines will be pruned.\n"
                    )

                    flg_last_level = True
                    break
            else:
                if self.timer.child_out_of_time:
                    logger.info(
                        "Time limit exceeded in one of the tasks. AutoML will blend level {0} models.\n".format(
                            leven_number
                        )
                    )
                    flg_last_level = True

            logger.info("\x1b[1mLayer {} training completed.\x1b[0m\n".format(leven_number))

            # here is split on exit condition
            if not flg_last_level:

                self.levels.append(pipes)
                level_predictions = concatenate(level_predictions)

                if self.skip_conn:
                    valid_part = train_valid.get_validation_data()
                    try:
                        # convert to initial dataset type
                        level_predictions = valid_part.from_dataset(level_predictions)
                    except TypeError:
                        raise TypeError(
                            "Can not convert prediction dataset type to input features. Set skip_conn=False"
                        )
                    level_predictions = concatenate([level_predictions, valid_part])
                train_valid = create_validation_iterator(level_predictions, None, n_folds=None, cv_iter=cv_iter)
            else:
                break

        blended_prediction, last_pipes = self.blender.fit_predict(level_predictions, pipes, self.targets_order)
        self.levels.append(last_pipes)

        self.reader.upd_used_features(remove=list(set(self.reader.used_features) - set(self.collect_used_feats())))

        del self._levels

        # saving automl model with joblib
        if path_to_save is not None:
            # There is 1 parameter for model save:
            # "path_to_save" - name of model for saving.

            dump_name = path_to_save if path_to_save.endswith(".joblib") else f"{path_to_save}.joblib"
            dump(self, dump_name, compress=0)

        if self.return_all_predictions:
            return concatenate(level_predictions)
        return blended_prediction

    def predict(
        self,
        data: Any,
        features_names: Optional[Sequence[str]] = None,
        return_all_predictions: Optional[bool] = None,
    ) -> LAMLDataset:
        """Predict with automl on new dataset.

        Args:
            data: Dataset to perform inference.
            features_names: Optional features names,
                if cannot be inferred from `train_data`.
            return_all_predictions: if True,
                returns all model predictions from last level

        Returns:
            Dataset with predictions.

        """
        dataset = self.reader.read(data, features_names=features_names, add_array_attrs=False)

        for n, level in enumerate(self.levels, 1):
            # check if last level

            level_predictions = []
            for _n, ml_pipe in enumerate(level):
                level_predictions.append(ml_pipe.predict(dataset))

            if n != len(self.levels):

                level_predictions = concatenate(level_predictions)

                if self.skip_conn:

                    try:
                        # convert to initial  dataset type
                        level_predictions = dataset.from_dataset(level_predictions)
                    except TypeError:
                        raise TypeError(
                            "Can not convert prediction dataset type to input features. Set skip_conn=False"
                        )
                    dataset = concatenate([level_predictions, dataset])
                else:
                    dataset = level_predictions
            else:
                if (return_all_predictions is None and self.return_all_predictions) or return_all_predictions:
                    return concatenate(level_predictions)
                return self.blender.predict(level_predictions)

    def collect_used_feats(self) -> List[str]:
        """Get feats that automl uses on inference.

        Returns:
            Features names list.

        """
        used_feats = set()

        for lvl in self.levels:
            for pipe in lvl:
                used_feats.update(pipe.used_features)

        used_feats = list(used_feats)

        return used_feats

    def collect_model_stats(self) -> Dict[str, int]:
        """Collect info about models in automl.

        Returns:
            Dict with models and its runtime numbers.

        """
        model_stats = {}

        for lvl in self.levels:
            for pipe in lvl:
                for ml_algo in pipe.ml_algos:
                    model_stats[ml_algo.name] = len(ml_algo.models)

        return model_stats



"""Main pytorch training and prediction class with Snapshots Ensemble."""

import logging

from copy import deepcopy
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import torch
import torch.nn as nn

from torch.utils.data import DataLoader
from tqdm import tqdm

from ..utils.logging import get_stdout_level
from .dp_utils import CustomDataParallel


try:
    from apex import amp
except:
    amp = None


from .utils import _dtypes_mapping


logger = logging.getLogger(__name__)


def optim_to_device(optim: torch.optim.Optimizer, device: torch.device) -> torch.optim.Optimizer:
    """Change optimizer device.

    Args:
        optim: Optimizer.
        device: To device.

    Returns:
        Optimizer on selected device.

    """
    for state in optim.state.values():
        for k, v in state.items():
            if torch.is_tensor(v):
                state[k] = v.to(device)
    return optim


class SnapshotEns:
    """In memory snapshots class.

    Class for SE, SWA and early stopping.

    Args:
        device: Torch device.
        k: Number of snapshots / checkpoint for swa.
        early_stopping: Use early stopping.
        patience: Patience before early stopping.
        swa: Use stochastic weight averaging.

    """

    def __init__(
        self,
        device: torch.device,
        k: int = 1,
        early_stopping: bool = True,
        patience: int = 3,
        swa: bool = False,
    ):
        self.best_loss = np.array([np.inf] * k)
        self.k = k
        self.device = device
        self.models = [nn.Module()] * k
        self.early_stopping = early_stopping
        self.patience = patience
        self.swa = swa
        self.counter = 0
        self.early_stop = False

    def update(self, model: nn.Module, loss: float):
        """Update current state.

        Args:
            model: Torch model.
            loss: Loss value, lower is better.

        """
        if np.any(self.best_loss > loss):
            self._sort()
            pos = np.where(self.best_loss > loss)[0][-1]

            self.best_loss[pos] = loss
            self.models[pos] = deepcopy(model.eval()).cpu()
            self.counter = 0
        else:
            self.counter += 1
            if self.counter >= self.patience:
                self.early_stop = True and self.early_stopping

    def predict(self, data: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Make snapshots prediction.

        Final prediction is a result of averaging predictions from snapshots.

        Args:
            data: Dict with model data.

        Returns:
            Torch tensor snapshot ensemble prediction.

        """
        preds = 0
        for model in self.models:
            model.eval().to(self.device)
            preds += model.predict(data)
            model.cpu()

        preds /= self.k
        return preds

    def forward(self, data: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Make snapshots forward , compute loss.

        Final result is a result of averaging losses from snapshots.

        Args:
            data: Dict with model data.

        Returns:
            Torch tensor snapshot ensemble loss.

        """
        preds = 0
        for model in self.models:
            model.eval().to(self.device)
            preds += model.forward(data)
            model.cpu()
        preds /= self.k
        return preds

    def _sort(self):
        """Sort models by loss value."""
        ids = np.argsort(self.best_loss)
        self.best_loss = self.best_loss[ids]
        self.models = [self.models[z] for z in ids]

    def set_weights(self, model: nn.Module, best: bool = False):
        """Set model weights as SWA or from best state.

        Args:
            model: Torch model.
            best: Save only best model.

        """
        n = 1 if best else min(self.k, sum(self.best_loss != np.inf))
        state_dict = {}
        for pos, score in enumerate(self.best_loss):

            if pos == n:
                break

            w = 1 / n

            new_state = self.models[pos].state_dict()
            # upd new state with weights
            for i in new_state.keys():
                new_state[i] = new_state[i].double() * w

            if pos == 0:
                state_dict = new_state
            else:
                # upd state
                for i in state_dict.keys():
                    state_dict[i] += new_state[i]

        model.load_state_dict(state_dict)

    def set_best_params(self, model: nn.Module):
        """Set best model params and clean cache.

        Args:
            model: Torch model.

        """
        self._sort()
        self.set_weights(model, best=False if self.swa else True)

        # TODO: think about dropping all models if use SWA. Change state_dict and load_state_dict
        # drop empty slots
        min_k = min(self.k, sum(self.best_loss != np.inf))
        self.models = self.models[:min_k]
        self.best_loss = self.best_loss[:min_k]

    def state_dict(self) -> Dict:
        """State SE dict.

        Returns:
            Dict with SE state.

        """
        models_dict = {"best_loss": self.best_loss}

        for n, model in enumerate(self.models):
            if isinstance(model, CustomDataParallel):
                models_dict[n] = model.module.state_dict()
            else:
                models_dict[n] = model.state_dict()
        return models_dict

    def load_state_dict(self, weights: Dict, model: nn.Module):
        """Load SE state.

        Args:
            model: Torch model.
            weights: State dict with weights.

        Returns:
            self.

        """
        self.best_loss = weights.pop("best_loss")
        self.models = [nn.Module()] * len(self.best_loss)
        for key, model_weights in weights.items():
            if isinstance(model, CustomDataParallel):
                model.module.load_state_dict(model_weights)
            else:
                model.load_state_dict(model_weights)
            self.models[key] = deepcopy(model.eval()).cpu()
        weights["best_loss"] = self.best_loss
        return self


class Trainer:
    """Torch main trainer class.

    Train, validation and test loops for NN models.

    Use DataParallel if device_ids is not None.

    Args:
        net: Uninitialized torch model.
        net_params: Dict with model params.
        opt: Uninitialized torch optimizer.
        opt_params: Dict with optim params.
        n_epochs: Number of training epochs.
        device: Torch device.
        device_ids: Ids of used gpu devices or None.
        metric: Callable metric for validation.
        snap_params: Dict with SE parameters.
        is_snap: Use snapshots.
        sch: Uninitialized torch scheduler.
        scheduler_params: Dict with scheduler params.
        verbose: Verbose every N epochs.
        verbose_inside: Number of steps between verbose
            inside epoch or None.
        verbose_bar: Show progress bar for each epoch
            during batchwise training.
        apex: Use apex (lead to GPU memory leak among folds).
        pretrained_path: Path to the pretrained model weights.
        stop_by_metric: es and scheduler will stop by metric.

    """

    def __init__(
        self,
        net,
        net_params: Dict,
        opt,
        opt_params: Dict,
        n_epochs: int,
        device: torch.device,
        device_ids: List[int],
        metric: Callable,
        snap_params: Dict,
        is_snap: bool = False,
        sch: Optional[Any] = None,
        scheduler_params: Optional[Dict] = None,
        verbose: int = 1,
        verbose_inside: Optional[int] = None,
        verbose_bar: bool = False,
        apex: bool = False,
        pretrained_path: Optional[str] = None,
        stop_by_metric: bool = False,
        clip_grad: bool = False,
        clip_grad_params: Optional[Dict] = None,
        **kwargs
    ):
        self.net = net
        self.net_params = net_params
        self.opt = opt
        self.opt_params = opt_params
        self.n_epochs = n_epochs
        self.device = device
        self.device_ids = device_ids
        self.is_snap = is_snap
        self.snap_params = snap_params
        self.sch = sch
        self.scheduler_params = scheduler_params if scheduler_params is not None else {}
        self.verbose = verbose
        self.metric = metric
        self.verbose_inside = verbose_inside
        self.verbose_bar = verbose_bar
        self.apex = apex
        self.pretrained_path = pretrained_path
        self.stop_by_metric = stop_by_metric
        self.clip_grad = clip_grad
        self.clip_grad_params = clip_grad_params if clip_grad_params is not None else {}

        self.dataloader = None
        self.model = None
        self.optimizer = None
        self.scheduler = None
        self.se = None
        self.amp = None
        self.is_fitted = False

    def clean(self):
        """Clean all models."""
        self.model = None
        self.optimizer = None
        self.scheduler = None
        self.se = None
        self.amp = None
        return self

    def _init(self):
        """Init all models."""
        self.model = self.net(**self.net_params)
        if self.device_ids is not None:
            self.model = CustomDataParallel(self.model, device_ids=self.device_ids)

        self.se = SnapshotEns(self.device, **self.snap_params)
        self.optimizer = self.opt(self.model.parameters(), **self.opt_params)
        self.amp = amp if self.apex else None
        if self.amp is not None:
            opt_level = "O1"
            self.model, self.optimizer = self.amp.initialize(self.model, self.optimizer, opt_level=opt_level)
        self.model.to(self.device)
        self.scheduler = self.sch(self.optimizer, **self.scheduler_params) if self.sch is not None else None
        return self

    def load_state(self, path: Union[str, Dict]):
        """Load all models state.

        Args:
            path: Path to state dict or state dict.

        Returns:
            self.

        """
        if isinstance(path, str):
            checkpoint = torch.load(path, map_location=torch.device("cpu"))
            self.pretrained_path = path
        else:
            checkpoint = path

        self._init()
        if checkpoint["se"] is not None:
            self.se.load_state_dict(checkpoint["se"], self.model)

        if isinstance(self.model, CustomDataParallel):
            self.model.module.load_state_dict(checkpoint["model"])
        else:
            self.model.load_state_dict(checkpoint["model"])

        self.optimizer.load_state_dict(checkpoint["optimizer"])
        self.optimizer = optim_to_device(self.optimizer, self.device)

        if checkpoint["amp"] is not None:
            self.amp.load_state_dict(checkpoint["amp"])

        self.is_fitted = True

        del checkpoint, path
        return self

    def state_dict(self, path: Optional[Any] = None):
        """Create all models state.

        Switch all models on cpu before saving state.

        Args:
            path: Path to save state dict.

        Returns:
            Checkpoint if path is not None.

        """
        self.model.to(torch.device("cpu"))
        self.optimizer = optim_to_device(self.optimizer, torch.device("cpu"))
        if isinstance(self.model, CustomDataParallel):
            model_checkpoint = self.model.module.state_dict()
        else:
            model_checkpoint = self.model.state_dict()

        checkpoint = {
            "model": model_checkpoint,
            "optimizer": self.optimizer.state_dict(),
            "amp": self.amp.state_dict() if self.apex else None,
            "se": self.se.state_dict() if self.is_snap else None,
        }
        if path is not None:
            torch.save(checkpoint, path)
            del checkpoint
            return self
        else:
            return checkpoint

    def fit(self, dataloaders: Dict[str, DataLoader]) -> np.ndarray:
        """Fit model.

        Args:
            dataloaders: Dict with torch dataloaders.

        Returns:
            Validation prediction.

        """
        if self.pretrained_path is not None:
            self.load_state(self.pretrained_path)
        elif (self.model is None) or (not self.is_fitted):
            self._init()

        train_log = []
        for epoch in range(self.n_epochs):
            self.epoch = epoch
            # train
            train_loss = self.train(dataloaders=dataloaders)
            train_log.extend(train_loss)
            # test
            val_loss, val_data, weights = self.test(dataloader=dataloaders["val"])
            if self.stop_by_metric:
                cond = -1 * self.metric(*val_data, weights)
            else:
                cond = np.mean(val_loss)
            self.se.update(self.model, cond)

            if self.sch is not None:
                self.scheduler.step(cond)

            if (self.verbose is not None) and ((epoch + 1) % self.verbose == 0):
                logger.info3(
                    "Epoch: {e}, train loss: {tl}, val loss: {vl}, val metric: {me}".format(
                        me=self.metric(*val_data, weights),
                        e=self.epoch,
                        tl=np.mean(train_loss),
                        vl=np.mean(val_loss),
                    )
                )
            if self.se.early_stop:
                break

        self.se.set_best_params(self.model)

        if self.is_snap:
            val_loss, val_data, weights = self.test(dataloader=dataloaders["val"], snap=True, stage="val")
            logger.info3(
                "Result SE, val loss: {vl}, val metric: {me}".format(
                    me=self.metric(*val_data, weights), vl=np.mean(val_loss)
                )
            )
        elif self.se.swa:
            val_loss, val_data, weights = self.test(dataloader=dataloaders["val"])
            logger.info3(
                "Early stopping: val loss: {vl}, val metric: {me}".format(
                    me=self.metric(*val_data, weights), vl=np.mean(val_loss)
                )
            )

        self.is_fitted = True

        return val_data[1]

    def train(self, dataloaders: Dict[str, DataLoader]) -> List[float]:
        """Training loop.

        Args:
            dataloaders: Dict with torch dataloaders.

        Returns:
            Loss.

        """
        loss_log = []
        self.model.train()
        running_loss = 0
        c = 0

        logging_level = get_stdout_level()
        if logging_level < logging.INFO and self.verbose and self.verbose_bar:
            loader = tqdm(dataloaders["train"], desc="train", disable=False)
        else:
            loader = dataloaders["train"]

        for sample in loader:
            data = {
                i: (sample[i].long().to(self.device) if _dtypes_mapping[i] == "long" else sample[i].to(self.device))
                for i in sample.keys()
            }

            loss = self.model(data).mean()
            if self.apex:
                with self.amp.scale_loss(loss, self.optimizer) as scaled_loss:
                    scaled_loss.backward()
            else:
                loss.backward()

            if self.clip_grad:
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), **self.clip_grad_params)
            self.optimizer.step()
            self.optimizer.zero_grad()

            loss = loss.data.cpu().numpy()
            loss_log.append(loss)
            running_loss += loss

            c += 1
            if self.verbose and self.verbose_bar and logging_level < logging.INFO:
                if self.verbose_inside and c % self.verbose_inside == 0:
                    val_loss, val_data, weights = self.test(dataloader=dataloaders["val"])
                    if self.stop_by_metric:
                        cond = -1 * self.metric(*val_data, weights)
                    else:
                        cond = np.mean(val_loss)
                    self.se.update(self.model, cond)

                    logger.info3(
                        "Epoch: {e}, iter: {c}, val loss: {vl}, val metric: {me}".format(
                            me=self.metric(*val_data, weights),
                            e=self.epoch,
                            c=c,
                            vl=np.mean(val_loss),
                        )
                    )
                loader.set_description("train (loss=%g)" % (running_loss / c))

        return loss_log

    def test(
        self, dataloader: DataLoader, stage: str = "val", snap: bool = False
    ) -> Tuple[List[float], Tuple[np.ndarray, np.ndarray]]:
        """Testing loop.

        Args:
            dataloader: Torch dataloader.
            stage: Train, val or test.
            snap: Use snapshots.

        Returns:
            Loss, (Target, OOF).

        """
        loss_log = []
        weights_log = []
        self.model.eval()
        pred = []
        target = []
        logging_level = get_stdout_level()
        if logging_level < logging.INFO and self.verbose and self.verbose_bar:
            loader = tqdm(dataloader, desc=stage, disable=False)
        else:
            loader = dataloader

        with torch.no_grad():
            for sample in loader:
                data = {
                    i: (sample[i].long().to(self.device) if _dtypes_mapping[i] == "long" else sample[i].to(self.device))
                    for i in sample.keys()
                }

                if snap:
                    output = self.se.predict(data)
                    loss = self.se.forward(data) if stage != "test" else None
                else:
                    output = self.model.predict(data)
                    loss = self.model(data) if stage != "test" else None

                if stage != "test":
                    loss = loss.mean().data.cpu().numpy()

                loss_log.append(loss)

                output = output.data.cpu().numpy()
                target_data = data["label"].data.cpu().numpy()
                weights = data.get("weight", None)
                if weights is not None:
                    weights = weights.data.cpu().numpy()

                pred.append(output)
                target.append(target_data)
                weights_log.extend(weights)

        self.model.train()

        return (
            loss_log,
            (
                np.vstack(target) if len(target[0].shape) == 2 else np.hstack(target),
                np.vstack(pred) if len(pred[0].shape) == 2 else np.hstack(pred),
            ),
            np.array(weights_log),
        )

    def predict(self, dataloader: DataLoader, stage: str) -> np.ndarray:
        """Predict model.

        Args:
            dataloader: Torch dataloader.
            stage: Train, val or test.

        Returns:
            Prediction.

        """
        loss, (target, pred), _ = self.test(stage=stage, snap=self.is_snap, dataloader=dataloader)
        return pred
